"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from functools import lru_cache

import numpy as np

from jax.core import JaxprEqn, ClosedJaxpr, Var, Jaxpr

from qrisp.jasp.jasp_expression.centerclass import Jaspr
from qrisp.jasp.primitives import AbstractQubit
from qrisp.jasp import TracingQuantumSession, OperationPrimitive


class ControlledJaspr(Jaspr):
    """
    This class enables the representation of controlled quantum functions.

    The advantage of treating controlled functions in a separate manner is
    that controlled function have an efficient control themselves:
    Performing an MCX gate into a newly allocated Qubit and controlling
    the function on that qubit.

    In addition to the Jaspr for representing itself, this class therefore
    also carries a Jaxpr for the base function.
    """

    __slots__ = ("base_jaspr", "ctrl_state")

    def __init__(self, base_jaspr, ctrl_state, stop_recursion = False):

        self.base_jaspr = base_jaspr
        self.ctrl_state = str(ctrl_state)

        if ctrl_state == "1" and base_jaspr.ctrl_jaspr is not None:
            controlled_jaspr = base_jaspr.ctrl_jaspr
        else:
            controlled_jaspr = multi_control_jaspr(
                base_jaspr, len(ctrl_state), ctrl_state
            )

        Jaspr.__init__(self, controlled_jaspr)
        self.envs_flattened = True
        
        if self.base_jaspr.inv_jaspr and not stop_recursion:
            self.inv_jaspr = ControlledJaspr(base_jaspr.inv_jaspr, ctrl_state, stop_recursion = True)
        

    def control(self, num_ctrl, ctrl_state=-1):

        if isinstance(ctrl_state, int):
            if ctrl_state < 0:
                ctrl_state += 2**num_ctrl

            ctrl_state = bin(ctrl_state)[2:].zfill(num_ctrl)
        else:
            ctrl_state = str(ctrl_state)

        return ControlledJaspr.from_cache(self.base_jaspr, ctrl_state + self.ctrl_state)

    def inverse(self):
        return ControlledJaspr.from_cache(self.base_jaspr.inverse(), self.ctrl_state)

    @classmethod
    @lru_cache(maxsize=int(1e5))
    def from_cache(cls, base_jaspr, ctrl_state):
        return ControlledJaspr(base_jaspr, ctrl_state)


control_var_count = np.zeros(1)


def copy_jaxpr(jaxpr):
    return Jaxpr(
        constvars=list(jaxpr.constvars),
        invars=list(jaxpr.invars),
        outvars=list(jaxpr.outvars),
        eqns=list(jaxpr.eqns),
        effects=jaxpr.effects,
    )


def control_eqn(eqn, ctrl_qubit_var):
    """
    Receives and equation that describes either an operation or a pjit primitive
    and returns an equation that describes the inverse.

    Parameters
    ----------
    eqn : jax.core.JaxprEqn
        The equation to be inverted.

    Returns
    -------
    inverted_eqn
        The equation with inverted operation.

    """
    from qrisp.jasp import Jaspr, AbstractQuantumCircuit

    if eqn.primitive.name == "pjit":

        new_params = dict(eqn.params)

        invars = list(eqn.invars)
        if isinstance(eqn.params["jaxpr"].jaxpr, Jaspr):
            new_params["jaxpr"] = ClosedJaxpr(
                new_params["jaxpr"].jaxpr.control(1), eqn.params["jaxpr"].consts
            )

            new_params["name"] = "c" + new_params["name"]

            invars = [ctrl_qubit_var] + eqn.invars

        return JaxprEqn(
            primitive=eqn.primitive,
            invars=invars,
            outvars=list(eqn.outvars),
            params=new_params,
            source_info=eqn.source_info,
            effects=eqn.effects,
        )

    elif eqn.primitive.name == "while":

        new_params = dict(eqn.params)
        body_jaxpr = eqn.params["body_jaxpr"].jaxpr
        cond_jaxpr = eqn.params["cond_jaxpr"].jaxpr

        if isinstance(
            body_jaxpr.invars[-1].aval, AbstractQuantumCircuit
        ) and isinstance(body_jaxpr.outvars[-1].aval, AbstractQuantumCircuit):

            # Generate controlled body jaxpr
            new_params["body_jaxpr"] = ClosedJaxpr(
                control_jaspr(Jaspr(eqn.params["body_jaxpr"].jaxpr)),
                eqn.params["body_jaxpr"].consts,
            )

            # Insert control value as output variable for the next iteration
            new_params["body_jaxpr"].jaxpr.outvars.insert(
                2, new_params["body_jaxpr"].jaxpr.invars[0]
            )
            new_params["body_jaxpr"].jaxpr.invars.insert(
                2, new_params["body_jaxpr"].jaxpr.invars.pop(0)
            )

        else:
            new_jaxpr = copy_jaxpr(new_params["body_jaxpr"].jaxpr)
            new_jaxpr.invars.insert(2, ctrl_qubit_var)
            new_params["body_jaxpr"] = ClosedJaxpr(
                new_jaxpr, eqn.params["body_jaxpr"].consts
            )

        if isinstance(
            cond_jaxpr.invars[-1].aval, AbstractQuantumCircuit
        ) and isinstance(cond_jaxpr.outvars[-1].aval, AbstractQuantumCircuit):
            new_params["cond_jaxpr"] = ClosedJaxpr(
                control_jaspr(Jaspr(eqn.params["cond_jaxpr"].jaxpr)),
                eqn.params["cond_jaxpr"].consts,
            )

            new_params["cond_jaxpr"].jaxpr.outvars.insert(
                2, new_params["cond_jaxpr"].jaxpr.invars[0]
            )
            new_params["cond_jaxpr"].jaxpr.invars.insert(
                2, new_params["cond_jaxpr"].jaxpr.invars.pop(0)
            )

        else:
            new_jaxpr = copy_jaxpr(new_params["cond_jaxpr"].jaxpr)
            new_jaxpr.invars.insert(2, ctrl_qubit_var)
            new_params["cond_jaxpr"] = ClosedJaxpr(
                new_jaxpr, eqn.params["cond_jaxpr"].consts
            )

        output_control_var = Var(suffix=str(control_var_count[0]), aval=AbstractQubit())
        control_var_count[0] += 1
        temp = JaxprEqn(
            primitive=eqn.primitive,
            invars=eqn.invars[:2] + [ctrl_qubit_var] + eqn.invars[2:],
            outvars=eqn.outvars[:2] + [output_control_var] + eqn.outvars[2:],
            params=new_params,
            source_info=eqn.source_info,
            effects=eqn.effects,
        )

        return temp
    elif eqn.primitive.name == "cond":

        new_params = dict(eqn.params)

        if isinstance(eqn.invars[-1].aval, AbstractQuantumCircuit) and isinstance(
            eqn.outvars[-1].aval, AbstractQuantumCircuit
        ):
            branch_list = []
            for i in range(len(new_params["branches"])):
                controlled_branch_jaxpr = control_jaspr(
                    Jaspr(new_params["branches"][i].jaxpr)
                )
                controlled_branch_jaxpr = ClosedJaxpr(
                    controlled_branch_jaxpr, new_params["branches"][i].consts
                )
                branch_list.append(controlled_branch_jaxpr)
        else:
            return eqn

        new_params["branches"] = tuple(branch_list)

        temp = JaxprEqn(
            primitive=eqn.primitive,
            invars=[eqn.invars[0], ctrl_qubit_var] + eqn.invars[1:],
            outvars=eqn.outvars,
            params=new_params,
            source_info=eqn.source_info,
            effects=eqn.effects,
        )
        return temp

    else:
        num_qubits = eqn.primitive.op.num_qubits
        return JaxprEqn(
            primitive=eqn.primitive.control(),
            invars=[ctrl_qubit_var]
            + eqn.invars[:-num_qubits]
            + eqn.invars[-num_qubits:],
            outvars=eqn.outvars,
            params=eqn.params,
            source_info=eqn.source_info,
            effects=eqn.effects,
        )


@lru_cache(int(1e5))
def control_jaspr(jaspr):
    """
    Takes a Jaspr and returns a Jaspr that has an additional Qubit argument
    (located behind the QuantumCircuit argument). The returned Jaspr is
    controlled on that Qubit argument.

    Parameters
    ----------
    Jaspr : qrisp.jasp.Jaspr
        The Jaspr to control.

    Returns
    -------
    controlled_jaspr : qrisp.jasp.Jaspr
        The controlled Jaspr.

    """

    from qrisp.jasp import Jaspr, AbstractQubit

    ctrl_qubit_var = Var(suffix=str(control_var_count[0]), aval=AbstractQubit())
    control_var_count[0] += 1

    new_eqns = []
    for eqn in jaspr.eqns:
        if isinstance(eqn.primitive, OperationPrimitive) or eqn.primitive.name in [
            "pjit",
            "while",
            "cond",
        ]:
            new_eqns.append(control_eqn(eqn, ctrl_qubit_var))
        elif eqn.primitive.name == "measure":
            raise Exception("Tried to applied quantum control to a measurement")
        else:
            new_eqns.append(eqn)

    permeability = dict(jaspr.permeability)
    permeability[ctrl_qubit_var] = True

    return Jaspr(
        permeability=permeability,
        isqfree=jaspr.isqfree,
        constvars=jaspr.constvars,
        invars=[ctrl_qubit_var] + jaspr.invars,
        outvars=jaspr.outvars,
        eqns=new_eqns,
    )


def multi_control_jaspr(jaspr, num_ctrl, ctrl_state):
    """
    Similar to control_jaspr but allows specification of more than
    one control and a control state

    Parameters
    ----------
    jaspr : qrisp.jasp.Jaspr
        The Jaspr to control.
    num_ctrl : int
        The amount of controls to add.
    ctrl_state : str
        The bitstring describing the control state.

    Returns
    -------
    qrisp.jasp.Jaspr
        The controlled Jaspr.

    """

    from qrisp.jasp import AbstractQubit, make_jaspr

    ctrl_vars = [
        Var(suffix=str(control_var_count[0] + _), aval=AbstractQubit())
        for _ in range(num_ctrl)
    ]
    control_var_count[0] += num_ctrl
    ctrl_avals = [x.aval for x in ctrl_vars]

    return make_jaspr(exec_multi_controlled_jaspr(jaspr, num_ctrl, ctrl_state))(
        *(ctrl_avals + [var.aval for var in jaspr.invars[:-1] + jaspr.constvars])
    )


def exec_multi_controlled_jaspr(jaspr, num_ctrls, ctrl_state):

    def multi_controlled_jaspr_executor(*args):

        qs = TracingQuantumSession.get_instance()
        ctrls = list(args)[:num_ctrls]
        invalues = list(args)[num_ctrls:]
        controlled_jaspr = control_jaspr(jaspr)

        from qrisp.circuit import XGate

        if num_ctrls == 1:

            if ctrl_state == "0":
                qs.append(XGate(), ctrls)
            temp = controlled_jaspr.embedd(*args, inline=True)
            if ctrl_state == "0":
                qs.append(XGate(), ctrls)
            return temp

        else:

            from qrisp import QuantumBool

            mcx_operation = XGate().control(num_ctrls, ctrl_state=ctrl_state)

            ctrl_qbl = QuantumBool(name="ctrl_qbl*")
            ctrl_qb = ctrl_qbl[0]

            qs.append(mcx_operation, ctrls + [ctrl_qb])
            res = controlled_jaspr.embedd(*([ctrl_qb] + invalues), inline=True)
            qs.append(mcx_operation, ctrls + [ctrl_qb])

            ctrl_qbl.delete()

            return res

    return multi_controlled_jaspr_executor



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from qrisp import *
import numpy as np
import jax.numpy as jnp

from numba import njit, prange
from jax import jit, vmap


def maxcut_obj(x, G):
    return maxcut_obj_jitted(int(x[::-1], 2), list(G.edges()))


@njit(cache=True)
def maxcut_obj_jitted(x, edge_list):
    cut = 0
    for i, j in edge_list:
        # the edge is cut
        if ((x >> i) ^ (x >> j)) & 1:
            # if x[i] != x[j]:
            cut -= 1
    return cut


@njit(parallel=True, cache=True)
def maxcut_energy(outcome_array, count_array, edge_list):

    res_array = np.zeros(len(outcome_array))
    for i in prange(len(outcome_array)):
        res_array[i] = maxcut_obj_jitted(outcome_array[i], edge_list) * count_array[i]

    return np.sum(res_array)


def create_maxcut_cl_cost_function(G):
    """
    Creates the classical cost function for an instance of the maximum cut problem for a given graph ``G``.

    Parameters
    ----------
    G : nx.Graph
        The Graph for the problem instance.

    Returns
    -------
    cl_cost_function : function
        The classical cost function for the problem instance, which takes a dictionary of measurement results as input.

    """

    def cl_cost_function(counts):

        edge_list = np.array(list(G.edges()), dtype=np.uint32)

        counts_keys = list(counts.keys())

        int_list = []
        if not isinstance(counts_keys[0], str):

            for c_array in counts_keys:
                integer = int("".join([c for c in c_array])[::-1], 2)
                int_list.append(integer)
        else:
            for c_str in counts_keys:
                integer = int(c_str[::-1], 2)
                int_list.append(integer)

        counts_array = np.array(list(counts.values()))
        outcome_array = np.array(int_list, dtype=np.uint32)

        return maxcut_energy(outcome_array, counts_array, edge_list)

    return cl_cost_function


@jit
def extract_boolean_digit(integer, digit):
    return (integer >> digit) & 1


def create_cut_computer(G):
    edge_list = jnp.array(G.edges())

    @jit
    def cut_computer(x):
        x_uint = jnp.uint32(x)
        bools = extract_boolean_digit(x_uint, edge_list[:, 0]) != extract_boolean_digit(
            x_uint, edge_list[:, 1]
        )
        cut = jnp.sum(bools)  # Count the number of edges crossing the cut
        return -cut

    return cut_computer


def create_maxcut_sample_array_post_processor(G):
    cut_computer = create_cut_computer(G)

    def post_processor(sample_array):
        cut_values = vmap(cut_computer)(
            sample_array
        )  # Use vmap for automatic vectorization
        average_cut = jnp.mean(cut_values)
        return average_cut

    return post_processor


def create_maxcut_cost_operator(G):
    r"""
    Creates the cost operator for an instance of the maximum cut problem for a given graph ``G``.

    Parameters
    ----------
    G : nx.Graph
        The Graph for the problem instance.

    Returns
    -------
    cost_operator : function
        A function receiving a :ref:`QuantumVariable` and a real parameter $\gamma$.
        This function performs the application of the cost operator.

    """

    def maxcut_cost_operator(qv, gamma):

        if not check_for_tracing_mode():
            if len(G) != len(qv):
                raise Exception(
                    f"Tried to call MaxCut cost Operator for graph of size {len(G)} on argument of invalid size {len(qv)}"
                )

        for pair in list(G.edges()):
            rzz(2 * gamma, qv[pair[0]], qv[pair[1]])
            # cx(qv[pair[0]], qv[pair[1]])
            # rz(2 * gamma, qv[pair[1]])
            # cx(qv[pair[0]], qv[pair[1]])
            # barrier(qv)

    return maxcut_cost_operator


def maxcut_problem(G):
    """
    Creates a QAOA problem instance with appropriate phase separator, mixer, and
    classical cost function.

    Parameters
    ----------
    G : nx.Graph
        The graph for the problem instance.

    Returns
    -------
    :ref:`QAOAProblem`
        A QAOA problem instance for MaxCut for a given graph ``G``.

    """
    from qrisp.qaoa import QAOAProblem, RX_mixer

    return QAOAProblem(
        create_maxcut_cost_operator(G), RX_mixer, create_maxcut_cl_cost_function(G)
    )



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from qrisp.core import mcx, h, s, s_dg, x
from qrisp.qtypes import QuantumBool
from qrisp.environments import invert


# Algorithm based on https://link.springer.com/article/10.1007/s10773-017-3389-4
def yong_mcx(input_qubits, target, ancilla=None, ctrl_state=None):
    if isinstance(target, list):
        if len(target) != 1:
            raise Exception("Tried to execute yong algorithm with multiple targets")
        target = target[0]

    if ctrl_state is None:
        ctrl_state = len(input_qubits) * "1"

    from qrisp.alg_primitives.mcx_algs import gray_pt_mcx

    if len(input_qubits) == 2:
        if ancilla is None:
            mcx(input_qubits, target, method="gray", ctrl_state=ctrl_state)
        else:
            target.qs().append(
                gray_pt_mcx(2, ctrl_state=ctrl_state), input_qubits + [target]
            )
            # mcx(input_qubits, target, method = "gray_pt", ctrl_state = ctrl_state)
        return

    for i in range(len(input_qubits)):
        if ctrl_state[i] == "0":
            x(input_qubits[i])

    ancilla_allocated = False

    if ancilla is None:
        ancilla_allocated = True
        ancilla_bl = QuantumBool(name="yong_anc*")
        ancilla = ancilla_bl[0]

    n = len(input_qubits)

    partition_k_1 = input_qubits[n // 2 :]
    partition_k_2 = input_qubits[: n // 2]

    h(target)

    yong_mcx(input_qubits=partition_k_1, target=ancilla, ancilla=partition_k_2[-1])

    s(ancilla)

    yong_mcx(
        input_qubits=partition_k_2 + [target], target=ancilla, ancilla=partition_k_1[-1]
    )

    s_dg(ancilla)

    with invert():
        s(ancilla)

        yong_mcx(
            input_qubits=partition_k_2 + [target],
            target=ancilla,
            ancilla=partition_k_1[0],
        )

        s_dg(ancilla)

        yong_mcx(input_qubits=partition_k_1, target=ancilla, ancilla=partition_k_2[0])

    h(target)

    if ancilla_allocated:
        ancilla_bl.delete()

    for i in range(len(input_qubits)):
        if ctrl_state[i] == "0":
            x(input_qubits[i])



"""
********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from qrisp import gphase, rz, cx, conjugate, custom_control, x, y, z
from qrisp.operators.qubit.visualization import X_, Y_, Z_

from sympy import Symbol
import numpy as np

PAULI_TABLE = {
    ("I", "I"): ("I", 1),
    ("I", "X"): ("X", 1),
    ("I", "Y"): ("Y", 1),
    ("I", "Z"): ("Z", 1),
    ("I", "A"): ("A", 1),
    ("I", "C"): ("C", 1),
    ("I", "P0"): ("P0", 1),
    ("I", "P1"): ("P1", 1),
    ("X", "I"): ("X", 1),
    ("X", "X"): ("I", 1),
    ("X", "Y"): ("Z", 1j),
    ("X", "Z"): ("Y", (-0 - 1j)),
    ("X", "A"): ("P1", 1),
    ("X", "C"): ("P0", 1),
    ("X", "P0"): ("C", 1),
    ("X", "P1"): ("A", 1),
    ("Y", "I"): ("Y", 1),
    ("Y", "X"): ("Z", (-0 - 1j)),
    ("Y", "Y"): ("I", 1),
    ("Y", "Z"): ("X", 1j),
    ("Y", "A"): ("P1", 1j),
    ("Y", "C"): ("P0", (-0 - 1j)),
    ("Y", "P0"): ("C", 1j),
    ("Y", "P1"): ("A", (-0 - 1j)),
    ("Z", "I"): ("Z", 1),
    ("Z", "X"): ("Y", 1j),
    ("Z", "Y"): ("X", (-0 - 1j)),
    ("Z", "Z"): ("I", 1),
    ("Z", "A"): ("A", 1),
    ("Z", "C"): ("C", -1),
    ("Z", "P0"): ("P0", 1),
    ("Z", "P1"): ("P1", -1),
    ("A", "I"): ("A", 1),
    ("A", "X"): ("P0", 1),
    ("A", "Y"): ("P0", 1j),
    ("A", "Z"): ("A", -1),
    ("A", "A"): ("I", 0),
    ("A", "C"): ("P0", 1),
    ("A", "P0"): ("I", 0),
    ("A", "P1"): ("A", 1),
    ("C", "I"): ("C", 1),
    ("C", "X"): ("P1", 1),
    ("C", "Y"): ("P1", (-0 - 1j)),
    ("C", "Z"): ("C", 1),
    ("C", "A"): ("P1", 1),
    ("C", "C"): ("I", 0),
    ("C", "P0"): ("C", 1),
    ("C", "P1"): ("I", 0),
    ("P0", "I"): ("P0", 1),
    ("P0", "X"): ("A", 1),
    ("P0", "Y"): ("A", (-0 - 1j)),
    ("P0", "Z"): ("P0", 1),
    ("P0", "A"): ("A", 1),
    ("P0", "C"): ("I", 0),
    ("P0", "P0"): ("P0", 1),
    ("P0", "P1"): ("I", 0),
    ("P1", "I"): ("P1", 1),
    ("P1", "X"): ("C", 1),
    ("P1", "Y"): ("C", 1j),
    ("P1", "Z"): ("P1", -1),
    ("P1", "A"): ("I", 0),
    ("P1", "C"): ("C", 1),
    ("P1", "P0"): ("I", 0),
    ("P1", "P1"): ("P1", 1),
}

#
# QubitTerm
#


class QubitTerm:
    r""" """

    def __init__(self, factor_dict={}):
        self.factor_dict = dict(factor_dict)

        self.hash_value = hash(tuple(sorted(factor_dict.items(), key=lambda x: x[0])))

    def update(self, update_dict):
        self.factor_dict.update(update_dict)
        self.hash_value = hash(tuple(sorted(self.factor_dict.items())))

    def __hash__(self):
        return self.hash_value

    def __eq__(self, other):
        return self.hash_value == other.hash_value

    def copy(self):
        return QubitTerm(self.factor_dict.copy())

    def is_identity(self):
        return len(self.factor_dict) == 0

    def binary_representation(self, n):
        x_vector = np.zeros(n, dtype=int)
        z_vector = np.zeros(n, dtype=int)
        for index in range(n):
            curr_factor = self.factor_dict.get(index, "I")
            if curr_factor == "X":
                x_vector[index] = 1
            elif curr_factor == "Z":
                z_vector[index] = 1
            elif curr_factor == "Y":
                x_vector[index] = 1
                z_vector[index] = 1

        return x_vector, z_vector

    def serialize(self):
        # This function serializes the QubitTerm in a way that facilitates the
        # measurement post-processing. To learn more details about the reasoning
        # behind this Serialisation check QubitOperator.conjugation_circuit

        # The idea here is to serialize the operator via 3 integers.
        # These integers specify how the energy of a measurement sample should
        # be computed.
        # They are processed in the function "evaluate_observable".

        # 1. The Z-int: The binary representation of this integer has a 1 at
        # every digit, where there is a Pauli term in self.

        # 2. The and_int: The binary representation has a 1 at every digit, which
        # should participate in an AND evaluation. If the evaluation of the AND
        # does not return True, the energy of this measurement is 0.

        # 3. The ctrl_int: Thi binary representation has a 1 at every digit,
        # which should be flipped before evaluating the AND value.

        z_int = 0
        and_int = 0
        ctrl_int = 0
        last_ladder_factor = None
        factor_dict = self.factor_dict
        for i in factor_dict.keys():
            bit = 1 << i

            # Go through the cases and update the appropriate integers
            if factor_dict[i] in ["X", "Y", "Z"]:
                z_int |= bit
                continue
            elif factor_dict[i] == "C":
                ctrl_int |= bit
                last_ladder_factor = bit
                pass
            elif factor_dict[i] == "A":
                last_ladder_factor = bit
                pass
            elif factor_dict[i] == "P0":
                pass
            elif factor_dict[i] == "P1":
                ctrl_int |= bit
                pass
            else:
                continue

            and_int |= bit

        # The last ladder factor should not participate in the AND but should
        # instead be treated like a Z Operator (see QubitOperator.get_conjugation_circuit
        # to learn why).
        if last_ladder_factor is not None:
            and_int ^= last_ladder_factor
            z_int ^= last_ladder_factor

        # Returns a tuple, which contains the relevant integers and a boolean
        # indicating whether this term contains any ladder operators.

        return (z_int, and_int, ctrl_int, int(last_ladder_factor is not None))

    def to_pauli(self):

        from qrisp.operators import X, Y, Z

        res = 1

        for i, factor in self.factor_dict.items():
            if factor == "X":
                res *= X(i)
            elif factor == "Y":
                res *= Y(i)
            elif factor == "Z":
                res *= Z(i)
            elif factor == "A":
                res *= (X(i) + 1j * Y(i)) * 0.5
            elif factor == "C":
                res *= (X(i) - 1j * Y(i)) * 0.5
            elif factor == "P0":
                res *= (Z(i) + 1) * 0.5
            elif factor == "P1":
                res *= (Z(i) - 1) * (-0.5)

        return res

    def adjoint(self):
        new_factor_dict = {}
        for i, factor in self.factor_dict.items():
            if factor in ["X", "Y", "Z", "P0", "P1"]:
                new_factor_dict[i] = factor
            elif factor == "A":
                new_factor_dict[i] = "C"
            elif factor == "C":
                new_factor_dict[i] = "A"

        return QubitTerm(new_factor_dict)

    #
    # Simulation
    #
    @custom_control(static_argnums=0)
    def simulate(self, coeff, qv, ctrl=None):

        from qrisp import (
            h,
            cx,
            rz,
            mcp,
            conjugate,
            control,
            QuantumBool,
            mcx,
            x,
            p,
            s,
            QuantumEnvironment,
            gphase,
            QuantumVariable,
            find_qs,
        )
        from qrisp.operators import QubitOperator
        import numpy as np

        # If required, do change of basis. Change of basis here means, that
        # the quantum argument is conjugated with a function that makes the
        # Operator diagonal. Please refer to the comments in QubitOperator.change_of_basis
        # to learn how this works.
        for factor in self.factor_dict.values():
            if factor not in ["I", "Z", "P0", "P1"]:
                qubit_op = QubitOperator({self: coeff})
                with conjugate(qubit_op.change_of_basis)(qv) as diagonal_op:
                    for diagonal_term, coeff in diagonal_op.terms_dict.items():
                        if ctrl is None:
                            diagonal_term.simulate(coeff, qv)
                        else:
                            with control(ctrl):
                                diagonal_term.simulate(coeff, qv)
                return

        qs = find_qs(qv)

        # We group the term into 2 types:

        # 1. Z-indices, where a Z-Operator is applied
        # 2. Projector indices, i.e. P0 or P1

        # Contains the indices of Pauli Z
        Z_indices = []

        # Contains the indices of the projection Operators
        projector_indices = []
        # Contains the booleans, which specify whether P0 or P1
        projector_state = []

        factor_dict = self.factor_dict
        for i in factor_dict.keys():
            if factor_dict[i] == "Z":
                Z_indices.append(i)
            elif factor_dict[i] == "P0":
                projector_indices.append(i)
                projector_state.append(False)
            elif factor_dict[i] == "P1":
                projector_indices.append(i)
                projector_state.append(True)

        # Determine the control qubits and the control state
        projector_ctrl_state = 0
        projector_qubits = []

        if ctrl is not None:
            projector_qubits.append(ctrl)
            projector_ctrl_state = 1
            # projector_ctrl_state = "1"

        for i in range(len(projector_indices)):
            # projector_ctrl_state += str(int(projector_state[i]))
            projector_ctrl_state += int(projector_state[i]) * 2 ** (
                len(projector_qubits)
            )
            projector_qubits.append(qv[projector_indices[i]])

        # If no non-trivial indices are found, we perform a global phase
        # and are done.
        if len(Z_indices + projector_qubits) == 0:
            gphase(-coeff, qv[0])
            return

        # If there are only projectors, the circuit is a mcp gate
        elif len(Z_indices) == 0:

            # Perform the mcp
            if len(projector_qubits) == 1:
                if projector_ctrl_state == 0:
                    p(coeff, projector_qubits[0])
                    gphase(-coeff, projector_qubits[0])
                else:
                    p(-coeff, projector_qubits[0])
            else:
                mcp(
                    -coeff,
                    projector_qubits,
                    method="balauca",
                    ctrl_state=projector_ctrl_state,
                )

            return

        # For your reference, here is the circuit that that implements
        # exp(i*phi*H)
        # For
        # H = A(0)*C(1)*C(2)*Z(3)*X(4)

        #           ┌───┐                                                                  ┌───┐
        #     qv.0: ┤ X ├────────────■───────────────────────────────────■─────────────────┤ X ├
        #           └─┬─┘┌───┐       │                                   │            ┌───┐└─┬─┘
        #     qv.1: ──┼──┤ X ├───────o───────────────────────────────────o────────────┤ X ├──┼──
        #             │  └─┬─┘┌───┐  │  ┌───┐┌───┐┌──────────────┐┌───┐  │  ┌───┐┌───┐└─┬─┘  │
        #     qv.2: ──■────■──┤ H ├──┼──┤ X ├┤ X ├┤ Rz(-1.0*phi) ├┤ X ├──┼──┤ X ├┤ H ├──■────■──
        #                     └───┘  │  └─┬─┘└─┬─┘└──────┬───────┘└─┬─┘  │  └─┬─┘└───┘
        #     qv.3: ─────────────────┼────■────┼─────────┼──────────┼────┼────■─────────────────
        #           ┌───┐            │         │         │          │    │  ┌───┐
        #     qv.4: ┤ H ├────────────┼─────────■─────────┼──────────■────┼──┤ H ├───────────────
        #           └───┘          ┌─┴─┐                 │             ┌─┴─┐└───┘
        # hs_anc.0: ───────────────┤ X ├─────────────────■─────────────┤ X ├────────────────────
        #                          └───┘                               └───┘

        # To implement the ladder operators, we use the ideas from https://arxiv.org/abs/2310.12256

        # In it's essence the circuits described there are a conjugation with an inverse GHZ state
        # preparation and a multi controlled RZ gate.

        # Since we performed the change of basis however already at the beginning
        # of this function, we don't have to perform the GHZ preparation anymore.
        # We only need to consider the projectors and the Z-indices

        # To realize the behavior of the projector indices, not that the
        # Hamiltonian has the form

        # H = P1*H_red
        #   = |1><1|*H_red

        # Simulating this H is therefore the controlled version of H_red:

        # exp(itH) = |0><0| ID + |1><1| exp(itH_red)

        # We can therefore add the projector indices to the computation of
        # the control qubit. If the projector indices are in the wrong state,
        # no phase will be performed, which is precisely what we want

        flip_control_phase = False
        # If the term has no projectors,
        # we don't need a controlled RZ
        if len(projector_qubits) == 0:
            env = QuantumEnvironment()
            control_qubit_available = False
        elif len(projector_qubits) == 1:
            # If there is only one projector qubit, we can use this as control value
            hs_anc = projector_qubits[0]
            control_qubit_available = True
            if not projector_ctrl_state == 1:
                flip_control_phase = True
            env = QuantumEnvironment()
        else:
            # In any other case, we allocate an additional quantum bool,
            # perform a mcx to compute the control value.
            # To achieve the multi-controlled RZ behavior, we control the RZ
            # on that quantum bool.

            hs_anc_qbl = QuantumBool(qs=qs, name="hs_anc*")
            hs_anc = hs_anc_qbl[0]
            control_qubit_available = True

            # Compute the control value

            from qrisp.alg_primitives.mcx_algs import balauca_layer

            def semi_balauca_mcx(projector_qubits, target, ctrl_state, ancillae):

                reduction_qubits = list(projector_qubits)
                fresh_ancillae = list(ancillae)
                ctrl_list = [
                    str(int((ctrl_state >> i) % 2))
                    for i in range(len(reduction_qubits))
                ]

                while len(reduction_qubits) > 2:

                    k = len(reduction_qubits) // 2

                    balauca_layer(
                        reduction_qubits[: 2 * k],
                        fresh_ancillae[:k],
                        structure=k * [2],
                        ctrl_list=[ctrl_list[i] for i in range(2 * k)],
                        use_mcm=True,
                    )

                    reduction_qubits = reduction_qubits[2 * k :] + fresh_ancillae[:k]
                    ctrl_list = ctrl_list[2 * k :] + k * ["1"]
                    fresh_ancillae = fresh_ancillae[k:]

                mcx(
                    reduction_qubits,
                    target,
                    method="gidney",
                    ctrl_state="".join(ctrl_list),
                )

            balauca_ancillae = QuantumVariable(
                len(projector_qubits) - 1, qs=qs, name="balauca_ancilla*"
            )

            env = conjugate(semi_balauca_mcx)(
                projector_qubits,
                hs_anc,
                projector_ctrl_state,
                ancillae=[
                    balauca_ancillae[k] for k in range(len(projector_qubits) - 1)
                ],
            )

        # Perform the conjugation
        with env:

            # The following qubit will be the target of the RZ gate
            anchor_index = Z_indices[-1]

            # Before we execute the RZ, we need to deal with the Z indices.

            # By inspection we see that acting on qubit 1 & 2 with Z operators,
            # is equivalent to the term (-1)**(q_1 (+) q_2)
            # It therefore suffices to compute the parity value of all the passive
            # qubits and flip the sign of the coefficient based on the parity.

            # To this end we perform several CX gate on-to the anchor qubit.
            # Since the anchor qubit will receive an RZ gate, each bitflip will
            # induce a sign flip of the phase.

            # We achieve this by executing a conjugation with the following function
            def flip_anchor_qubit(qv, anchor_index, Z_indices):
                for i in Z_indices:
                    cx(qv[i], qv[anchor_index])

            # Perform the conjugation
            with conjugate(flip_anchor_qubit)(
                qv, anchor_index=anchor_index, Z_indices=Z_indices[:-1]
            ):

                # Perform the controlled RZ
                if control_qubit_available:
                    # Use Selinger's circuit (page 5)
                    with conjugate(cx)(qv[anchor_index], hs_anc):
                        rz(coeff, qv[anchor_index])
                        if flip_control_phase:
                            coeff = -coeff
                        rz(-coeff, hs_anc)
                else:
                    rz(coeff * 2, qv[anchor_index])

        if len(projector_indices) >= 2:
            # Delete ancilla
            hs_anc_qbl.delete()
            balauca_ancillae.delete()

    #
    # Printing
    #

    def __str__(self):
        # Convert the sympy expression to a string and return it
        expr = self.to_expr()
        return str(expr)

    def __repr__(self):
        return str(self)

    def non_trivial_indices(self):
        res = set()
        for index, P in self.factor_dict.items():
            res.add(index)
        return res

    def to_expr(self):
        """
        Returns a SymPy expression representing the QubitTerm.

        Returns
        -------
        expr : sympy.expr
            A SymPy expression representing the QubitTerm.

        """

        def to_spin(P, index):
            if P == "I":
                return 1
            if P == "X":
                return X_(index)
            if P == "Y":
                return Y_(index)
            if P == "Z":
                return Z_(index)
            if P == "A":
                return Symbol("A(" + str(index) + ")", commutative=False)
            if P == "C":
                return Symbol("C(" + str(index) + ")", commutative=False)
            if P == "P0":
                return Symbol("P0(" + str(index) + ")", commutative=False)
            if P == "P1":
                return Symbol("P1(" + str(index) + ")", commutative=False)

        expr = 1
        index_list = sorted(list(self.factor_dict.keys()))

        for i in index_list:
            expr = expr * to_spin(self.factor_dict[i], str(i))

        return expr

    #
    # Arithmetic
    #

    def __pow__(self, e):
        if isinstance(e, int) and e >= 0:
            if e % 2 == 0:
                return QubitTerm({(): 1})
            else:
                return self
        else:
            raise TypeError(
                "Unsupported operand type(s) for ** or pow(): "
                + str(type(self))
                + " and "
                + str(type(e))
            )

    def __mul__(self, other):
        result_factor_dict = {}
        result_coeff = 1
        a = self.factor_dict
        b = other.factor_dict

        keys = set(a.keys()) | set(b.keys())
        for key in keys:
            factor, coeff = PAULI_TABLE[a.get(key, "I"), b.get(key, "I")]
            if coeff == 0:
                return QubitTerm({}), 0
            if factor != "I":
                result_factor_dict[key] = factor
                result_coeff *= coeff
        return QubitTerm(result_factor_dict), result_coeff

    def subs(self, subs_dict):
        """

        Parameters
        ----------
        subs_dict : dict
            A dictionary with indices (int) as keys and numbers (int, float, complex) as values.

        Returns
        -------
        QubitTerm
            The resulting QubitTerm.
        result_coeff : int, float, complex
            The resulting coefficient.

        """
        result_factor_dict = self.factor_dict.copy()
        result_coeff = 1

        for key, value in subs_dict.items():
            if key in result_factor_dict:
                del result_factor_dict[key]
                result_coeff *= value

        return QubitTerm(result_factor_dict), result_coeff
    
    #
    # Unitary
    #

    def unitary(self, sign = False):
        """
        Returns the unitary for the term self. 

        Parameters
        ----------
            sign : bool
                A Boolean indicating whether a phase -1 must be applied. The default is False.
        
        """
        def unitary_(operand):
            for i, factor in self.factor_dict.items():
                if factor=="X":
                    x(operand[i])
                elif factor=="Y":
                    y(operand[i])
                elif factor=="Z":
                    z(operand[i])

            if sign:
                gphase(np.pi, operand[0])
            
        return unitary_
    
    def apply_unitary(self, operand, sign = False):
        """
        Applies the unitary for the term self to the ``operand``. 

        Parameters
        ----------
            sign : bool
                A Boolean indicating whether a phase -1 must be applied. The default is False.
        
        """
        for i, factor in self.factor_dict.items():
            if factor=="X":
                x(operand[i])
            elif factor=="Y":
                y(operand[i])
            elif factor=="Z":
                z(operand[i])

        if sign:
                gphase(np.pi, operand[0])
            
    #
    # Commutativity
    #
    def commutator(self, other):
        from qrisp.operators.qubit import QubitOperator

        term_0, coeff_0 = self * other
        term_1, coeff_1 = other * self

        temp = QubitOperator({term_0: coeff_0}) - QubitOperator({term_1: coeff_1})
        temp.apply_threshold(0.5)

        return temp

    def commute(self, other):
        """
        Checks if two QubitTerms commute.

        """
        a = self.factor_dict
        b = other.factor_dict

        keys = set()
        keys.update(set(a.keys()))
        keys.update(set(b.keys()))

        sign_flip = 1

        for key in keys:
            factor_a = PAULI_TABLE[a.get(key, "I"), b.get(key, "I")]
            factor_b = PAULI_TABLE[b.get(key, "I"), a.get(key, "I")]

            if factor_a[1] == 0 and factor_b[1] == 0:
                return True
            if factor_a[0] == factor_b[0]:
                if factor_a[1] == factor_b[1]:
                    continue
                elif factor_a[1] == -factor_b[1]:
                    sign_flip *= -1
                    continue
            else:
                return len(self.commutator(other).terms_dict) == 0

        return sign_flip == 1

    def commute_pauli(self, other):
        """
        Checks if the Pauli factors of two QubitTerms commute and the ladder factors commute qubit-wise.

        """
        a = self.factor_dict
        b = other.factor_dict

        keys = set()
        keys.update(set(a.keys()))
        keys.update(set(b.keys()))

        sign_flip = 1

        for key in keys:
            factor_a = PAULI_TABLE[a.get(key, "I"), b.get(key, "I")]
            factor_b = PAULI_TABLE[b.get(key, "I"), a.get(key, "I")]

            if not factor_a == factor_b:
                if factor_a[0] in ["I", "X", "Y", "Z"] and factor_b[0] in [
                    "I",
                    "X",
                    "Y",
                    "Z",
                ]:
                    if factor_a[1] == -factor_b[1]:
                        sign_flip *= -1
                else:
                    return False

        return sign_flip == 1

    def commute_qw(self, other):
        """
        Checks if two QubitTerms commute qubit-wise.

        """
        a = self.factor_dict
        b = other.factor_dict

        keys = set()
        keys.update(set(a.keys()))
        keys.update(set(b.keys()))

        for key in keys:
            if (
                not PAULI_TABLE[a.get(key, "I"), b.get(key, "I")]
                == PAULI_TABLE[b.get(key, "I"), a.get(key, "I")]
            ):
                return False
        return True

    def intersect(self, other):
        """
        Checks if two QubitTerms operate on the same qubit.

        """
        return (
            len(set(self.factor_dict.keys()).intersection(other.factor_dict.keys()))
            != 0
        )

    def ladders_agree(self, other):
        """
        Checks if the ladder operators of two QubitTerms operate on the same set of qubits.

        Parameters
        ----------
        other : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        ladder_indices_self = [
            factor[0] for factor in self.factor_dict.items() if factor[1] in ["A", "C"]
        ]
        ladder_indices_other = [
            factor[0] for factor in other.factor_dict.items() if factor[1] in ["A", "C"]
        ]
        return set(ladder_indices_self) == set(ladder_indices_other)

    def ladders_intersect(self, other):
        """
        Checks if the ladder operators of two QubitTerms operate on the same qubit.

        Parameters
        ----------
        other : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        ladder_indices_self = [
            factor[0] for factor in self.factor_dict.items() if factor[1] in ["A", "C"]
        ]
        ladder_indices_other = [
            factor[0] for factor in other.factor_dict.items() if factor[1] in ["A", "C"]
        ]
        return len(set(ladder_indices_self).intersection(ladder_indices_other)) != 0



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from jax.core import ClosedJaxpr, Literal
from jax import make_jaxpr
import jax.numpy as jnp
from qrisp.jasp import check_for_tracing_mode


class ContextDict(dict):

    def __getitem__(self, key):
        if isinstance(key, Literal):
            res = key.val
        else:
            res = dict.__getitem__(self, key)

        if type(res) == int:
            return jnp.array(res, dtype=jnp.dtype("int64"))
        elif type(res) == float:
            return jnp.array(res, dtype=jnp.dtype("float64"))
        else:
            return res


def exec_eqn(eqn, context_dic):
    invalues = extract_invalues(eqn, context_dic)
    res = eqn.primitive.bind(*invalues, **eqn.params)
    insert_outvalues(eqn, context_dic, res)


def eval_jaxpr(jaxpr, return_context_dic=False, eqn_evaluator=exec_eqn):
    """
    Evaluates a Jaxpr using the given context dic to replace variables

    Parameters
    ----------
    jaxpr : jax.core.Jaxpr
        The jaxpr to evaluate.

    Returns
    -------
    None.

    """

    from qrisp.jasp import Jaspr

    if isinstance(jaxpr, ClosedJaxpr):
        consts = list(jaxpr.consts)
        jaxpr = jaxpr.jaxpr
    elif isinstance(jaxpr, Jaspr):
        consts = list(jaxpr.consts)
    else:
        consts = []

    def jaxpr_evaluator(*args):

        args = consts + list(args)
        temp_var_list = jaxpr.constvars + jaxpr.invars

        if len(temp_var_list) != len(args):
            raise Exception("Tried to evaluate jaxpr with insufficient arguments")

        context_dic = ContextDict({temp_var_list[i]: args[i] for i in range(len(args))})
        eval_jaxpr_with_context_dic(jaxpr, context_dic, eqn_evaluator)

        if return_context_dic:
            outvals = [context_dic]
        else:
            outvals = []

        for i in range(len(jaxpr.outvars)):
            outvals.append(context_dic[jaxpr.outvars[i]])

        if len(outvals) == 1:
            return outvals[0]
        else:
            return tuple(outvals)

    return jaxpr_evaluator


def reinterpret(jaxpr, eqn_evaluator=exec_eqn):

    if isinstance(jaxpr, ClosedJaxpr):
        inter_jaxpr = jaxpr.jaxpr
    else:
        inter_jaxpr = jaxpr

    res = make_jaxpr(eval_jaxpr(inter_jaxpr, eqn_evaluator=eqn_evaluator))(
        *[var.aval for var in jaxpr.constvars + jaxpr.invars]
    ).jaxpr

    res.constvars.extend(res.invars[: len(inter_jaxpr.constvars)])
    temp = list(res.invars[len(inter_jaxpr.constvars) :])
    res.invars.clear()
    res.invars.extend(temp)

    if isinstance(jaxpr, ClosedJaxpr):
        res = ClosedJaxpr(res, jaxpr.consts)

    return res


def eval_jaxpr_with_context_dic(jaxpr, context_dic, eqn_evaluator=exec_eqn):

    # Iterate through the equations
    for eqn in jaxpr.eqns:
        # Evaluate the primitive
        default_eval = eqn_evaluator(eqn, context_dic)

        if default_eval:
            if (
                eqn.primitive.name in ["while", "cond", "scan"]
                and not check_for_tracing_mode()
            ):

                from qrisp.jasp import (
                    evaluate_cond_eqn,
                    evaluate_while_loop,
                    evaluate_scan,
                )

                if eqn.primitive.name == "while":
                    evaluate_while_loop(eqn, context_dic, eqn_evaluator)
                elif eqn.primitive.name == "cond":
                    evaluate_cond_eqn(eqn, context_dic, eqn_evaluator)
                else:
                    evaluate_scan(eqn, context_dic, eqn_evaluator)

                continue

            exec_eqn(eqn, context_dic)


def extract_invalues(eqn, context_dic):
    invalues = []
    for i in range(len(eqn.invars)):
        invar = eqn.invars[i]
        invalues.append(context_dic[invar])
    return invalues


def extract_constvalues(eqn, context_dic):
    constvalues = []
    for i in range(len(eqn.constvars)):
        constvar = eqn.constvars[i]
        constvalues.append(context_dic[constvar])

    return constvalues


def insert_outvalues(eqn, context_dic, outvalues):

    if eqn.primitive.multiple_results:
        for i in range(len(eqn.outvars)):
            context_dic[eqn.outvars[i]] = outvalues[i]
    else:
        context_dic[eqn.outvars[0]] = outvalues



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

# This file implements the carry generation art of the algorithm presented here:
# https://arxiv.org/abs/2304.02921

# For this we implement the Brent-Kung-Tree in a recursive manner. See:
# https://web.stanford.edu/class/archive/ee/ee371/ee371.1066/lectures/lect_04.pdf Page 13

import numpy as np

from qrisp.core.gate_application_functions import cx, mcx
from qrisp.qtypes import QuantumVariable, QuantumBool
from qrisp.circuit import QuantumCircuit
from qrisp.permeability import auto_uncompute
from qrisp.misc import check_if_fresh
from qrisp.alg_primitives.mcx_algs import hybrid_mcx


# Returns the PROPAGATE status of a group of entries
def calc_P_group(P):
    new_p = QuantumBool(name="p_group*", qs=P[0].qs())
    # Due to the semi-classical nature of the algorithm, it is possible
    # that some propagate values are known to be 0 (because some of the
    # input values are known to be 0)
    # The following loop checks if they are |0> and if so, the corresponding
    # mcx gate is skipped
    for p in P:
        if isinstance(p, QuantumBool):
            p = p[0]
        if check_if_fresh([p], p.qs()):
            return new_p
    if len(P) == 2:
        mcx(P, new_p, method="gidney")
    else:
        hybrid_mcx(P, new_p)
    return new_p[0]


# Returns the GENERATE status of a group of entries
# A group has GENERATE status True, if any of the members
# has GENERATE status True and all of the following members
# have PROPAGATE status True.


# The GENERATE status of the group is calculated into the last
# member of the group (G[-1])
def calc_G_group(P, G):

    for i in range(len(G) - 1):

        controls = [G[i]] + P[i + 1 :]
        for p in controls:
            if isinstance(p, QuantumBool):
                p = p[0]
            if check_if_fresh([p], p.qs()):
                break
        else:

            if len(P[i + 1 :]) == 1:
                mcx(controls, G[-1], method="jones")
            else:
                hybrid_mcx(controls, G[-1])

    return G[-1]


# Receives the values PROPAGATE and GENERATE values of a group.
# It is assumed that the 0-th entry of G already contains the carry status
# This function then calculates the carry status for every other entry of G
def propagate_carry(P, G):

    # This loop iteratively calculates the CARRY status into all entries
    # of G.

    # We calculate the CARRY status of position i by considering the CARRY status of
    # position 0 and the GENERATE status of any position above that. To successfully
    # generate a CARRY status of True, all the other positions in between need to have
    # PROPAGATE status True.

    # Note that in order to calculate the CARRY status of position i, we need the
    # GENERATE status (not the CARRY status) of j for all j < i.
    # This is why we reverse the loop, such that the position with the highest i
    # are calculated first (the computation can access all the GENERATE
    # information because none of the CARRY status has been calculated yet)

    for i in range(1, len(G))[::-1]:

        for j in range(i):
            if len(P[j + 1 : i + 1]) == 1:
                method = "jones"
            else:
                method = "auto"

            controls = [G[j]] + P[j + 1 : i + 1]
            for p in controls:
                if isinstance(p, QuantumBool):
                    p = p[0]
                if check_if_fresh([p], p.qs()):
                    break
            else:
                if len(P[j + 1 : i + 1]) == 1:
                    mcx(controls, G[i], method=method)
                else:
                    hybrid_mcx(controls, G[i])


# This function recursively calculates the layers of the Brent-Kung tree.
# It receives the GENERATE and CARRY status of a set of positions and returns
# the evaluated CARRY status

# The function takes the parameters P and G, which represent the PROPAGATE and
# the GENERATE status of the corresponding digit.

# The parameter r describes the radix of the Brent-Kung tree.

# Finally the parameter k_out describes a the layer where the carry calculation
# should be ended prematurely. That is if k_out = 2 and r = 4, the output will
# be every 4**2 = 16-th bit of G, that contains a properly calculated carry value


def brent_kung_tree(P, G, r, k_out=np.inf):

    from qrisp import t_depth_indicator

    gate_speed = lambda x: t_depth_indicator(x, 2**-6)

    # if len(G):
    # print(G[0].qs().compile(40, compile_mcm = True, gate_speed = gate_speed).t_depth())
    if r <= 1 or not isinstance(r, int):
        raise Exception("Provided invalid radix base (needs to be integer > 1)")

    # We group the positions and calculate the GENERATE and PROPAGATE status
    # of the groups. Subsequently the function is called recursively to evaluate the
    # CARRY status of the respective groups.

    # Finally the CARRY status of the groups is propagated to the remaining positions
    # to yield the complete set of carries (not only the groups)

    # Recursion cancellation condition
    if len(G) == 0:
        return []

    # Set up list for grouped GENERATE and PROPAGATE states
    grouped_P = []
    grouped_G = []
    for i in range(len(G) // r):
        interval = slice(r * i, r * (i + 1))

        if i == 0:
            # We can skip the 0-th position
            # This is because the propagate_carry function and the calc_G_group
            # function are not taking the 0-th position into consideration
            grouped_P.append(None)
        else:
            grouped_P.append(calc_P_group(P[interval]))

        grouped_G.append(calc_G_group(P[interval], G[interval]))

    # Call the function recursively
    grouped_carry = brent_kung_tree(grouped_P, grouped_G, r, k_out=k_out - 1)

    if k_out > 0:
        return grouped_carry

    # If working with a radix > 2, the carry also needs to be propagated
    # to the position of the first group. In the case of r = 2, this is not neccessary
    # because the first group consists just of the 0-th position (which already contains
    # the carry in its GENERATE status) and the 1st position which contains the CARRY status
    # due to the GENERATE computation of this group.
    initial_interval = slice(0, r - 1)
    propagate_carry(P[initial_interval], G[initial_interval])

    # Go through all the groups and propagate the carry.
    # Note that the last interval has not neccessarily size r (depending on the size of G)
    for i in range(1, len(G) // r + 1):
        interval = slice(r * i - 1, r * (i + 1) - 1)
        propagate_carry(P[interval], G[interval])

    # Return G (now contains the carry positions)
    return G


# The following function calculates the carry status from two QuantumVariables of equal size

# The function takes two QuantumFloats a, b, the radix base and the radix exponent.
# The function returns a QuantumVariable c, that contains the CARRY status of every
# radix_base**radix_exponent digit of CARRY(a,b)

# Example:

# a = 3 = 0011
# b = 6 = 0110
# CARRY(a,b) = 0110

# Assume we choose radix_base = 2. If we choose radix_exponent = 0,
# this function returns

# c = 0110

# If we choose radix_exponent = 1

# c = 01

# ie. position 1 and 3 of CARRY(a,b)

# If we chose radix_exponent = 2

# c = 0

# ie. position 3 of CARRY(a,b)


# Furthermore this function uncomputes all garbage. What is the garbage here?
# The cancelation of the brent kung tree at an early layer produce many GENERATE
# entries, that are not holding CARRY values. Furthermore, also the PROPAGATE values
# of groups are uncomputed.


@auto_uncompute
def qq_calc_carry(a, b, radix_base=2, radix_exponent=0):

    R = radix_base**radix_exponent
    # How can we achieve that the GENERATE entries, that don't contain a relevant
    # CARRY information are uncomputed while the CARRY entries stay?

    # We do this by creating two QuantumVariables, one which holds the carry and
    # one which holds the GENERATE garbage. We then arrange both of them in a list g,
    # that represents the GENERATE array.

    # This variable will hold the result
    # If b can be divided into k blocks of size R,
    # we only need k-1 ancillae qubit, because we have no need for
    # the carry of the last bock.

    c = QuantumVariable(int(np.ceil(len(b) / R)) - 1, name="carry*", qs=b[0].qs())

    # This variable will hold the intermediate GENERATE values, that are supposed
    # to be uncomputed. The uncomputation is performed using the auto_uncompute
    # decorator. This decorator uncomputes all local variables.
    if R > 1:
        brent_kung_ancilla = QuantumVariable(
            c.size * (R - 1), name="bk_ancilla*", qs=b[0].qs()
        )
        anc_list = list(brent_kung_ancilla)
    else:
        anc_list = []

    # Create the g list
    c_list = list(c)
    g = []
    for i in range(len(c)):
        for j in range(R - 1):
            g.append(anc_list.pop(0))
        g.append(c_list.pop(0))

    # Append the remaining ancillae
    g.extend(anc_list)

    # We now compute the GENERATE and the PROPAGATE status of each digit
    # g_i = a_i * b_i
    # p_i = a_i XOR b_i

    # This could be achieved by applying a sequence of mcx gates to compute g
    # followed by a sequence of cx gates that compute p.

    # This however breaks the automatic uncomputation algorithm since it induces
    # a cycle in the dag representation. We can still use automatic uncomputation
    # with a trick: We apply the CNOT gate first (instead of the mcx gate) and create
    # a ancillary gate that wraps the mcx gates within two CNOT gates.
    # This ancillary gate is then permeable on the "b_i" input and the cycle in the
    # DAG dissapears. This introduces two extra CNOT gates per iteration. However,
    # two of these CNOT gates are adjacent so, they will be cancelled out during
    # circuit optimization.

    # This gate is synthesized by the anc_gate function.
    for i in range(min(len(g), len(a), len(b))):
        cx(a[i], b[i])
        b[0].qs().append(anc_gate(), [a[i], b[i], g[i]])

    p = b

    # Calculate compute the carry using the Brent-Kung tree
    brent_kung_tree(p, g, radix_base, radix_exponent)

    # Restore the original b value.
    for i in range(min(len(g), len(a), len(b))):
        cx(a[i], b[i])

    # Return the result
    return c


# This function returns a gate, that performs the toffoli and cx gate.
# We mark the second qubit as permeable to "hack" the uncomputation algorithm.
def anc_gate():
    from qrisp.alg_primitives.mcx_algs import GidneyLogicalAND

    anc_qc = QuantumCircuit(3)
    qubits = anc_qc.qubits
    anc_qc.cx(qubits[0], qubits[1])
    anc_qc.append(GidneyLogicalAND(), [qubits[0], qubits[1], qubits[2]])
    anc_qc.cx(qubits[0], qubits[1])
    res = anc_qc.to_gate(name="anc_gate")
    res.permeability = {0: True, 1: True, 2: False}
    res.is_qfree = True
    return res



"""
********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from qrisp import QuantumArray, mcp, conjugate, invert
from qrisp.jasp import q_fori_loop, q_cond, check_for_tracing_mode
from jax import lax
import sympy as sp
import numpy as np
import jax.numpy as jnp


def QITE(qarg, U_0, exp_H, s, k, method="GC"):
    r"""
    Performs `Double-Braket Quantum Imaginary-Time Evolution (DB-QITE) <https://arxiv.org/abs/2412.04554>`_.
    Given a Hamiltonian :ref:`Operator <Operators>` $H$, this method implements the unitary $U_k$ that is recursively defined by either of

    * Group commutator (GQ) approximation:

    .. math::

        U_{k+1} = e^{i\sqrt{s_k}H}e^{i\sqrt{s_k}\omega_k}e^{-i\sqrt{s_k}H}U_k

    * Higher-order product formula (HOPF) approximation:

    .. math::

        U_{k+1} = e^{i\phi\sqrt{s_k}H}e^{i\phi\sqrt{s_k}\omega_k}e^{-i\sqrt{s_k}H}e^{-i(1+\phi)\sqrt{s_k}\omega_k}e^{i(1-\phi)\sqrt{s_k}H}U_k

    where $e^{-it\omega_k}=U_ke^{it\ket{0}\bra{0}}U_k^{\dagger}$ is the refection around the state $\ket{\omega_k}=U_k\ket{0}$.



    Parameters
    ----------
    qarg : :ref:`QuantumVariable` or :ref:`QuantumArray`
        The quantum argument on which quantum imaginary time evolution is performed.
    U_0 : function
        A Python function that takes a QuantumVariable or QuantumArray ``qarg`` as input, and prepares the initial state.
    exp_H : function
        A Python function that takes a QuantumVariable or QuantumArray ``qarg`` and time ``t`` as input, and performs forward evolution $e^{-itH}$.
    s : list[float] or list[Sympy.Symbol]
        A list of evolution times for each step.
    k : int
        The number of steps.
    method : str, optional
        The method for approximating the double-bracket flow (DBF). Available are ``GC`` and ``HOPF``.
        The default is ``GC``.


    Examples
    --------

    We utilize QITE to approximate the ground state energy of a Heisenberg chain. We start by defining the lattice graph $G$:

    ::

        import networkx as nx

        # Create a graph
        N = 4
        G = nx.Graph()
        G.add_edges_from([(k,k+1) for k in range(N-1)])

    Next, we set up the Heisenberg Hamiltonian and calculate the ground state energy classically:

    ::

        from qrisp.operators import X, Y, Z

        def create_heisenberg_hamiltonian(G):
            H = sum(X(i)*X(j)+Y(i)*Y(j)+Z(i)*Z(j) for (i,j) in G.edges())
            return H

        H = create_heisenberg_hamiltonian(G)
        print(H)
        print(H.ground_state_energy())


    As explained :ref:`in this example <VQEHeisenberg>`, a suitable initial approximation for the ground state is given by a tensor product of singlet states $\frac{1}{\sqrt{2}}(\ket{10}-\ket{01})$ corresponding to a maximal matching of the graph $G$.
    Accordingly, we define the function ``U_0``:

    ::

        from qrisp import QuantumVariable
        from qrisp.vqe.problems.heisenberg import create_heisenberg_init_function

        M = nx.maximal_matching(G)
        U_0 = create_heisenberg_init_function(M)

        def state_prep():
            qv = QuantumVariable(N)
            U_0(qv)
            return qv

        E_0 = H.expectation_value(state_prep)()
        print(E_0)

    For the function ``exp_H`` that performs forward evolution $e^{-itH}$, we use the :meth:`trotterization <qrisp.operators.qubit.QubitOperator.trotterization>` method with 5 Trotter steps:

    ::

        def exp_H(qv, t):
            H.trotterization(method='commuting')(qv,t,5)

    With all the necessary ingredients, we use QITE to approximate the ground state:

    ::

        import numpy as np
        import sympy as sp
        from qrisp.qite import QITE

        steps = 4
        s_values = np.linspace(.01,.3,10)

        theta = sp.Symbol('theta')
        optimal_s = [theta]
        optimal_energies = [E_0]

        for k in range(1,steps+1):

            # Perform k steps of QITE
            def state_prep():
                qv = QuantumVariable(N)
                QITE(qv, U_0, exp_H, optimal_s, k)
                return qv

            qv = state_prep()
            qc = qv.qs.compile()

            # Find optimal evolution time
            # Use "precompliled_qc" keyword argument to avoid repeated compilation of the QITE circuit
            energies = [H.expectation_value(state_prep, diagonalisation_method='commuting', subs_dic={theta:s_}, precompiled_qc=qc)() for s_ in s_values]

            index = np.argmin(energies)
            s_min = s_values[index]

            optimal_s.insert(-1,s_min)
            optimal_energies.append(energies[index])

        print(optimal_energies)

    Finally, we visualize the results:

    ::

        import matplotlib.pyplot as plt

        evolution_times = [sum(optimal_s[i] for i in range(k)) for k in range(steps+1)]

        plt.xlabel('Evolution time', fontsize=15, color='#444444')
        plt.ylabel('Energy', fontsize=15, color='#444444')
        plt.axhline(y=H.ground_state_energy(), color='#6929C4', linestyle='--', linewidth=2, label='Exact energy')
        plt.plot(evolution_times, optimal_energies, c='#20306f', marker="o", linestyle='solid', linewidth=3, zorder=3, label='DB-QITE')
        plt.legend(fontsize=12, labelcolor='linecolor')
        plt.tick_params(axis='both', labelsize=12)
        plt.grid()
        plt.show()

    .. figure:: /_static/heisenberg_qite.png
        :scale: 80%
        :align: center

    """
    if not check_for_tracing_mode():

        if k == 0:
            U_0(qarg)
        else:
            s_ = sp.sqrt(s[k - 1])

            def conjugator(qarg):
                with invert():
                    QITE(qarg, U_0, exp_H, s, k - 1, method=method)

            def reflection(qarg, t_):
                with conjugate(conjugator)(qarg):
                    if isinstance(qarg,QuantumArray):
                        qubits = sum([qv.reg for qv in qarg.flatten()], [])
                        mcp(t_, qubits, ctrl_state=0, method="khattar")
                    else:
                        mcp(t_, qarg, ctrl_state=0, method="khattar")

            if method == "GC":

                QITE(qarg, U_0, exp_H, s, k - 1, method=method)

                with conjugate(exp_H)(qarg, s_):
                    reflection(qarg, s_)

            if method == "HOPF":

                phi = (sp.sqrt(5) - 1) / 2

                QITE(qarg, U_0, exp_H, s, k - 1, method=method)

                # exp_H performs forward evolution $e^{-itH}
                exp_H(qarg, -(1 - phi) * s_)
                reflection(qarg, -(1 + phi) * s_)
                exp_H(qarg, s_)
                reflection(qarg, phi * s_)
                exp_H(qarg, -phi * s_)

    else:

        """
        To create a jasp-compatible implementation of QITE, we need to remove the recursive structure.
        We achieve this by fully expanding the recursive formula for $U_k$ down to the $k=0$ level.
        From there, we find a tree structure with branching factor 3 (GC) or 5 (HOPF) where
        some branches are inverted due to the presence of conjugate operators $U_i^\dagger$.
        We traverse the tree depth-first using up-, down-, bounce-, and leaf-operations that we obtain from
        inspecting the formula for $U_k$.
        """

        def int_to_base(n, base=3, max_digits=10):
            """
            Get the array representation of an integer `n` with base `base`.
            The array has length `max_digits` and the least significant digit is at index `0`.
            """

            def cond_fun(state):
                n, digits, i = state
                return jnp.logical_and(n > 0, i < max_digits)

            def body_fun(state):
                n, digits, i = state
                digits = digits.at[i].set(n % base)
                n = n // base
                return n, digits, i + 1

            init_digits = jnp.zeros((max_digits,), dtype=jnp.int32)
            _, digits, _ = lax.while_loop(cond_fun, body_fun, (n, init_digits, 0))
            return digits

        # Define basic operations
        def U_0_dag(q_arg):
            with invert():
                U_0(q_arg)

        def exp_00(q_arg, time):
            mcp(time, q_arg, ctrl_state=0)

        if method == "GC":

            def body_fun(i, val):
                qarg = val

                # Obtain old and new position
                old_pos = int_to_base(i, 3)
                new_pos = int_to_base(i + 1, 3)
                # Obtain largest changed index + 1
                num_changes = jnp.count_nonzero(new_pos != old_pos)

                # Compute which operations must be inverted
                inv_mode_leaf = jnp.equal(jnp.count_nonzero(old_pos == 1) % 2, 0)
                inv_mode_up = inv_mode_leaf
                inv_mode_bounce = jnp.logical_xor(
                    inv_mode_up, old_pos[num_changes - 1] == 1
                )
                inv_mode_down = jnp.equal(jnp.count_nonzero(new_pos == 1) % 2, 0)

                # Apply U_0
                q_cond(inv_mode_leaf, U_0, U_0_dag, qarg)

                # Go up the branch
                time = q_fori_loop(
                    0, num_changes - 1, lambda j, time: time + jnp.sqrt(s[j]), 0
                )
                q_cond(inv_mode_up, exp_H, lambda a, b: None, qarg, -time)

                # Bounce to next branch
                q_cond(
                    jnp.logical_and(old_pos[num_changes - 1] == 0, inv_mode_bounce),
                    exp_H,
                    lambda a, b: None,
                    qarg,
                    jnp.sqrt(s[num_changes - 1]),
                )
                q_cond(
                    jnp.logical_and(old_pos[num_changes - 1] == 1, inv_mode_bounce),
                    exp_00,
                    lambda a, b: None,
                    qarg,
                    jnp.sqrt(s[num_changes - 1]),
                )
                q_cond(
                    jnp.logical_and(
                        old_pos[num_changes - 1] == 0, jnp.logical_not(inv_mode_bounce)
                    ),
                    exp_00,
                    lambda a, b: None,
                    qarg,
                    -jnp.sqrt(s[num_changes - 1]),
                )
                q_cond(
                    jnp.logical_and(
                        old_pos[num_changes - 1] == 1, jnp.logical_not(inv_mode_bounce)
                    ),
                    exp_H,
                    lambda a, b: None,
                    qarg,
                    -jnp.sqrt(s[num_changes - 1]),
                )

                # Go down to leaf
                q_cond(inv_mode_down, lambda a, b: None, exp_H, qarg, time)

                return qarg

            # Iterate all leafs except last
            q_fori_loop(0, 3**k - 1, body_fun, qarg)

            # Do last leaf
            U_0(qarg)
            time = lax.fori_loop(0, k, lambda j, time: time + jnp.sqrt(s[j]), 0)
            exp_H(qarg, -time)

        if method == "HOPF":

            phi = (jnp.sqrt(5) - 1) / 2

            def body_fun(i, val):
                qarg = val

                # Obtain old and new position
                old_pos = int_to_base(i, 5)
                new_pos = int_to_base(i + 1, 5)
                # Obtain largest changed index + 1
                num_changes = jnp.count_nonzero(new_pos != old_pos)

                # Compute which operations must be inverted
                inv_mode_leaf = (
                    jnp.count_nonzero(old_pos == 1) + jnp.count_nonzero(old_pos == 3)
                ) % 2 == 0
                inv_mode_up = inv_mode_leaf
                inv_mode_bounce = jnp.logical_xor(
                    inv_mode_up,
                    jnp.logical_or(
                        old_pos[num_changes - 1] == 1, old_pos[num_changes - 1] == 3
                    ),
                )
                inv_mode_down = (
                    jnp.count_nonzero(new_pos == 1) + jnp.count_nonzero(new_pos == 3)
                ) % 2 == 0

                # Apply U_0
                q_cond(inv_mode_leaf, U_0, U_0_dag, qarg)

                # Go up the branch
                time = (
                    q_fori_loop(
                        0, num_changes - 1, lambda j, time: time + jnp.sqrt(s[j]), 0
                    )
                    * phi
                )
                q_cond(inv_mode_up, exp_H, lambda a, b: None, qarg, -time)

                # Bounce to next branch
                q_cond(
                    jnp.logical_and(old_pos[num_changes - 1] == 0, inv_mode_bounce),
                    exp_H,
                    lambda a, b: None,
                    qarg,
                    -(1 - phi) * jnp.sqrt(s[num_changes - 1]),
                )
                q_cond(
                    jnp.logical_and(old_pos[num_changes - 1] == 1, inv_mode_bounce),
                    exp_00,
                    lambda a, b: None,
                    qarg,
                    -(1 + phi) * jnp.sqrt(s[num_changes - 1]),
                )
                q_cond(
                    jnp.logical_and(old_pos[num_changes - 1] == 2, inv_mode_bounce),
                    exp_H,
                    lambda a, b: None,
                    qarg,
                    jnp.sqrt(s[num_changes - 1]),
                )
                q_cond(
                    jnp.logical_and(old_pos[num_changes - 1] == 3, inv_mode_bounce),
                    exp_00,
                    lambda a, b: None,
                    qarg,
                    phi * jnp.sqrt(s[num_changes - 1]),
                )

                q_cond(
                    jnp.logical_and(
                        old_pos[num_changes - 1] == 3, jnp.logical_not(inv_mode_bounce)
                    ),
                    exp_H,
                    lambda a, b: None,
                    qarg,
                    (1 - phi) * jnp.sqrt(s[num_changes - 1]),
                )
                q_cond(
                    jnp.logical_and(
                        old_pos[num_changes - 1] == 2, jnp.logical_not(inv_mode_bounce)
                    ),
                    exp_00,
                    lambda a, b: None,
                    qarg,
                    (1 + phi) * jnp.sqrt(s[num_changes - 1]),
                )
                q_cond(
                    jnp.logical_and(
                        old_pos[num_changes - 1] == 1, jnp.logical_not(inv_mode_bounce)
                    ),
                    exp_H,
                    lambda a, b: None,
                    qarg,
                    -jnp.sqrt(s[num_changes - 1]),
                )
                q_cond(
                    jnp.logical_and(
                        old_pos[num_changes - 1] == 0, jnp.logical_not(inv_mode_bounce)
                    ),
                    exp_00,
                    lambda a, b: None,
                    qarg,
                    -phi * jnp.sqrt(s[num_changes - 1]),
                )

                # Go down to leaf
                q_cond(inv_mode_down, lambda a, b: None, exp_H, qarg, time)

                return qarg

            # Iterate all leafs except last
            q_fori_loop(0, 5**k - 1, body_fun, qarg)

            # Do last leaf
            U_0(qarg)
            time = -phi * lax.fori_loop(0, k, lambda j, time: time + jnp.sqrt(s[j]), 0)
            exp_H(qarg, time)



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import networkx as nx

from qrisp.permeability.qc_transformations.memory_management import topological_sort

# This function reorders the circuit such that the intended measurements can be executed
# as early as possible. Additionally, any instructions that are not needed for the
# intended measurements are removed

# Intended measurements has to be a list of qubits

# The strategy is similar to the one presented in reorder_circuit:
# We bring the circuit in the dag representation and perform a topological sort with
# the intended measurement as prefered instructions


# After that we investigate the circuit for instructions that can be removed
def lightcone_reduction(qc, intended_measurements):
    qc = qc.copy()

    # Insert intended measurements into the circuit
    for qb in intended_measurements:
        qc.measure(qb)

    # Generate dag representation
    from qrisp.permeability import PermeabilityGraph, TerminatorNode

    G = PermeabilityGraph(qc, remove_artificials=True)

    # Create result qc
    qc_new = qc.clearcopy()

    # Define prefered instructions
    measure_identifier = (
        lambda x: x.op.name == "measure" and x.qubits[0] in intended_measurements
    )

    def sub_sort(dag):
        nodes = list(dag.nodes())

        def sort_key(x):
            if isinstance(x, TerminatorNode):
                return 0
            else:
                return x.qc_index

        nodes.sort(key=sort_key)
        return nodes

    # Perform topological sort
    for n in topological_sort(G, prefer=measure_identifier, sub_sort=sub_sort):
        if n.instr:
            qc_new.append(n.instr)

    # Check which instructions come after the final measurement
    for i in range(len(qc_new.data))[::-1]:
        if measure_identifier(qc_new.data[i]):
            break

    redundant_qc = qc_new.clearcopy()

    redundant_qc.data = qc_new.data[i + 1 :]

    G = PermeabilityGraph(redundant_qc, remove_artificials=True)

    # #Now we need to make sure we don't remove deallocation gates from the data
    # #because this would inflate the qubit count of the compiled circuit

    # #The strategy here is that if we find a deallocation gate
    # #we remove any instruction involving the deallocated qubit from the list
    # #redundant instructions.

    for node in G.nodes():
        if node.instr is None:
            continue

        if node.instr.op.name == "qb_dealloc":
            ancs = nx.ancestors(G, node)

            redundant_qc.data.remove(node.instr)
            # print(f"removed {node.instr}")
            for pred in ancs:
                try:
                    redundant_qc.data.remove(pred.instr)
                except ValueError:
                    pass

    redundant_instructions = redundant_qc.data

    # We now remove the redundant instructions and the inserted
    # measurements from the circuit data
    i = 0
    while i < len(qc_new.data):
        if measure_identifier(qc_new.data[i]):
            qc_new.data.pop(i)
            continue
        if qc_new.data[i] in redundant_instructions:
            qc_new.data.pop(i)
            continue
        i += 1

    return qc_new



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from jax import make_jaxpr
from jax.core import ClosedJaxpr
from qrisp.jasp import flatten_environments, flatten_pjit, eval_jaxpr, make_jaspr


def jasp_function_test(func):

    def testing_function(*args):

        qv = func(*args)
        from qrisp.core import QuantumVariable

        qv.__class__ = QuantumVariable

        old_counts_dic = qv.get_measurement()

        jaspr = make_jaspr(func)(*args)

        qv_qubits, qc = jaspr.to_qc(*args)

        clbit_list = []
        for qb in qv_qubits:
            clbit_list.append(qc.measure(qb))

        counts = qc.run(shots=None)

        # Remove other measurements outcomes from counts dic
        new_counts_dic = {}
        for key in counts.keys():
            # Remove possible whitespaces
            new_key = key.replace(" ", "")
            # Remove other measurements
            new_key = new_key[: len(clbit_list)][::-1]

            # new_key = int(new_key, base=2)
            try:
                new_counts_dic[new_key] += counts[key]
            except KeyError:
                new_counts_dic[new_key] = counts[key]

        for k in old_counts_dic.keys():
            if abs(old_counts_dic[k] - new_counts_dic[k]) > 1e-4:
                return False
        return True

    return testing_function



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from jax.tree_util import tree_flatten, tree_unflatten
from qrisp.jasp.jasp_expression import make_jaspr


def qjit(function):
    """
    Decorator to leverage the jasp + Catalyst infrastructure to compile the given
    function to QIR and run it on the Catalyst QIR runtime.

    Parameters
    ----------
    function : callable
        A function performing Qrisp code.

    Returns
    -------
    callable
        A function executing the compiled code.

    Examples
    --------

    We write a simple function using the QuantumFloat quantum type and execute
    via ``qjit``:

    ::

        from qrisp import *
        from qrisp.jasp import qjit

        @qjit
        def test_fun(i):
            qv = QuantumFloat(i, -2)
            with invert():
                cx(qv[0], qv[qv.size-1])
                h(qv[0])
            meas_res = measure(qv)
            return meas_res + 3


    We execute the function a couple of times to demonstrate the randomness

    >>> test_fun(4)
    [array(5.25, dtype=float64)]
    >>> test_fun(5)
    [array(3., dtype=float64)]
    >>> test_fun(5)
    [array(7.25, dtype=float64)]

    """

    def jitted_function(*args):

        if not hasattr(function, "jaspr_dict"):
            function.jaspr_dict = {}

        args = list(args)

        signature = tuple([type(arg) for arg in args])
        if not signature in function.jaspr_dict:
            function.jaspr_dict[signature] = make_jaspr(function)(*args)

        return function.jaspr_dict[signature].qjit(
            *args, function_name=function.__name__
        )

    return jitted_function




#from qrisp.circuit.standard_operations import op_list
from qrisp.circuit.standard_operations import     XGate,YGate, ZGate,    CXGate,  CYGate,    CZGate,MCXGate,PGate,  CPGate,u3Gate,  HGate,RXGate,   RYGate,   RZGate,   MCRXGate,SGate , TGate, RXXGate,RZZGate,  SXGate,   SXDGGate,  Barrier, Measurement,  Reset,  QubitAlloc, QubitDealloc,GPhaseGate,  SwapGate,U1Gate,  IDGate
import numpy as np
from qrisp import QuantumVariable

def pytket_rand_test():
    qvRand = QuantumVariable(10)
    qcRand = qvRand.qs
    rotation = np.pi/2
    single_gates = [    
        XGate(),
        YGate(),
        ZGate(),
        HGate(),
        SXGate(), 
        SGate()]

    rot_gates = [
        RXGate,
        RYGate,
        RZGate,
        PGate]

    c_gates = [
        CXGate(),
        CYGate(),
        CZGate()]

    mc_gates = [
            SwapGate(),     
            ]

    mc_rot_gates =[
        RXXGate,
        RZZGate
            ]

    special_gates = [
                    MCXGate(control_amount=3)
                    ]


    op_list = [*mc_gates, *c_gates, *single_gates, *rot_gates, 
                *special_gates,
                ]

    used_ops = []
    for index in range(30):
        randInteg = np.random.randint(0,len(op_list)+1)
        if randInteg == len(op_list): 
            used_ops.append("mcrxxx")
        else:
            used_ops.append(op_list[randInteg])
        

    for op in used_ops:
        qubit_1 = qvRand[np.random.randint(7,9)]
        qubit_2 = qvRand[np.random.randint(3,6)]

        if op == "mcrxxx": 
            qcRand.append(MCRXGate(rotation, control_amount=3),[qvRand[0],qvRand[2],qubit_2,qubit_1])
        elif op in single_gates:
            qcRand.append(op, qubit_1)
        elif op in rot_gates:
            qcRand.append(op(rotation), qubit_1)
        # this is being called first due to mcx is subclass of cx reasons
        elif op in special_gates:
            qcRand.append(op, [qvRand[0],qvRand[2],qubit_2,qubit_1])
        elif op in mc_rot_gates:
        #elif op in c_gates:
            qcRand.append(op(rotation), [qubit_1,qubit_2])
        elif op in c_gates or mc_gates:
        #elif op in c_gates:
            qcRand.append(op, [qubit_1,qubit_2])

    #from qrisp.interface.converter.PyTket.convert_from_tket import convert_to_tket



    tket_qcRand = qcRand.to_pytket()
    #tket_qcRand = convert_to_tket(qc=qcRand)
    from pytket.extensions.qiskit import AerBackend
 
    tket_qcRand.measure_all()


    backend = AerBackend()
    if not backend.valid_circuit(tket_qcRand):
        compiled_circ = backend.get_compiled_circuit(tket_qcRand)
        assert backend.valid_circuit(compiled_circ)
    else: 
        compiled_circ = backend.get_compiled_circuit(tket_qcRand)
    #compiled_circ = tket_qcRand
    handle = backend.process_circuit(compiled_circ, n_shots=1000)
    result = backend.get_result(handle)
    cnt = result.get_counts()
    d = {}
    for key, value in cnt.items():
        Key = (str(index) for index in key)
        converted2 = ''.join(Key)
        #reverse string because of pytket specific return??
        #converted = converted2[::-1]
        converted = converted2
        d.setdefault(converted,value/1000)

    print(d)

    #print(result.get_counts(basis=BasisOrder.dlo)) 
    theRes = qvRand.get_measurement()
    print(theRes)

    for index4 in list(d.keys()):
        if not index4 in list(theRes.keys()):
            print("NOT IN THERE")
            print(index4)
        assert index4 in list(theRes.keys())
        if not theRes[index4]*0.6 <= d[index4] <= theRes[index4]*1.4:
            print(index4)
            print(theRes[index4])
            print(d[index4])
        if not d[index4] < 0.05:
            assert theRes[index4]*0.6 <= d[index4] <= theRes[index4]*1.4
    




"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from jax.core import AbstractValue, raise_to_shaped_mappings
from qrisp.jasp.primitives import QuantumPrimitive, AbstractQubitArray


class AbstractQuantumCircuit(AbstractValue):

    def __repr__(self):
        return "QuantumCircuit"

    def __hash__(self):
        return hash(AbstractQuantumCircuit)

    def __eq__(self, other):
        return isinstance(other, AbstractQuantumCircuit)


def create_qubits(size, state):
    return create_qubits_p.bind(size, state)


raise_to_shaped_mappings[AbstractQuantumCircuit] = lambda aval, _: aval

# Register Creation
create_qubits_p = QuantumPrimitive("create_qubits")
create_qubits_p.multiple_results = True


@create_qubits_p.def_abstract_eval
def create_qubits_abstract_eval(size, qc):
    """Abstract evaluation of the primitive.

    This function does not need to be JAX traceable. It will be invoked with
    abstractions of the actual arguments.
    Args:
      xs, ys, zs: abstractions of the arguments.
    Result:
      a ShapedArray for the result of the primitive.
    """

    assert isinstance(qc, AbstractQuantumCircuit)

    return AbstractQubitArray(), AbstractQuantumCircuit()


@create_qubits_p.def_impl
def create_qubit_impl(size, qc):
    from qrisp.circuit import QubitAlloc

    qubit_list = []

    for i in range(int(size)):
        qubit_list.append(qc.add_qubit())
        qc.append(QubitAlloc(), [qubit_list[-1]])

    return qubit_list, qc


# Register Deletion
delete_qubits_p = QuantumPrimitive("delete_qubits")


@delete_qubits_p.def_abstract_eval
def delete_qubits_abstract_eval(qarr, qc):
    """Abstract evaluation of the primitive.

    This function does not need to be JAX traceable. It will be invoked with
    abstractions of the actual arguments.
    Args:
      xs, ys, zs: abstractions of the arguments.
    Result:
      a ShapedArray for the result of the primitive.
    """

    return AbstractQuantumCircuit()


@delete_qubits_p.def_impl
def delete_qubits_impl(qarr, qc):
    from qrisp.circuit import QubitDealloc

    for i in range(len(qarr)):
        qc.append(QubitDealloc(), [qarr[i]])

    return qc


quantum_kernel_p = QuantumPrimitive("quantum_kernel")


@quantum_kernel_p.def_abstract_eval
def quantum_kernel_abstract_eval():
    """Abstract evaluation of the primitive.

    This function does not need to be JAX traceable. It will be invoked with
    abstractions of the actual arguments.
    Args:
      xs, ys, zs: abstractions of the arguments.
    Result:
      a ShapedArray for the result of the primitive.
    """

    return AbstractQuantumCircuit()



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

# Created by manzanillo
import pytest
from unittest.mock import patch
from qrisp.interface import IQMBackend, VirtualBackend

import pytest
from unittest.mock import patch, MagicMock
from qrisp.interface import IQMBackend

# Mock the IQMProvider and transpile_to_IQM functions from qiskit_iqm
def test_IQMBackend():
    # TO DO: set up proper mock backend
    return
    # Mock the server_url and token
    api_token = "mock_api_token"

    # Create a mock device instance
    device_instance = "garnet:fake"

    from iqm.iqm_client import CircuitCompilationOptions
    compilation_options = CircuitCompilationOptions()
    # Create an instance of IQMBackend
    backend = IQMBackend(api_token, device_instance, compilation_options = compilation_options)

    # Check that the backend is an instance of VirtualBackend
    assert isinstance(backend, VirtualBackend)

    # Create a mock QASM string
    from qrisp import QuantumCircuit
    qc = QuantumCircuit(4, name = "test")

    # Call the run method
    result = backend.run(qc, shots=1000, token=api_token)


# Test that the API token is a string
def test_IQMBackend_api_token_string_is_missing():
    with pytest.raises(TypeError):
        IQMBackend(123, "garnet")

# Test that the device instance is a string
def test_IQMBackend_device_instance_string_is_missing():
    with pytest.raises(TypeError):
        IQMBackend("mock_api_token", 123)


"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from qrisp.interface import BackendClient

api_endpoint = "127.0.0.1"


def CirqSim():
    return BackendClient(api_endpoint, port=8083)


def PennylaneSim():
    return BackendClient(api_endpoint, port=8084)


def MQTSim():
    return BackendClient(api_endpoint, port=8085)


def PennylaneRigettiSim():
    return BackendClient(api_endpoint, port=8086)


def PyTketStimSim():
    return BackendClient(api_endpoint, port=8087)


def QulacsSim():
    return BackendClient(api_endpoint, port=8088)


def QSimCirq():
    return BackendClient(api_endpoint, port=8089)


def QiskitSim():
    return BackendClient(api_endpoint, port=8090)


def QiboSim():
    return BackendClient(api_endpoint, port=8091)


def QiskitMPS():
    return BackendClient(api_endpoint, port=8092)



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import networkx as nx

from qrisp.circuit import fast_append, ControlledOperation, PTControlledOperation
from qrisp.permeability.type_checker import is_qfree

from qrisp.permeability.permeability_dag import (
    PermeabilityGraph,
    InstructionNode,
    TerminatorNode,
    AllocNode,
)


def uncompute_qc(qc, uncomp_qbs, recompute_qubits=[]):
    """
    This function applies the Unqomp algorithm to the QuantumCircuit qc, uncomputing
    the qubits uncomp_qbs. It is possible to specify a set of qubits that has been
    uncomputed earlier, which is allowed to be recomputed.

    Parameters
    ----------
    qc : QuantumCircuit
        The QuantumCircuit to uncompute.
    uncomp_qbs : list[Qubit]
        A list of Qubits to uncompute using the Unqomp algorithm.
    recompute_qubits : list[Qubit], optional
        A list of Qubits, that have been uncomputed earlier, which the algorithm
        is allowed to recompute. The default is [].

    Raises
    ------
    Exception
        Cyclic dependency detected in DAG during uncomputation.
    Exception
        Uncomputation failed because gate needs to be uncomputed
        but is also targeting qubits which are not up for uncomputation

    Returns
    -------
    uncomputed_qc
        The result of the Unqomp algorithm.

    """

    with fast_append():

        # To speed up the uncomputation, the first step is to filter
        # out instruction that are guaranteed that they don't need to be uncomputed.

        # The idea is that every instruction before the first instruction that operators
        # on the uncomputation qubits satisfies this property.
        # Same for every instruction after the last instruction.
        qc_new = qc.copy()
        previous_instructions = []
        follow_up_instructions = []

        # Determine the the previous instructions.
        while len(qc_new.data):
            instr = qc_new.data[0]
            if set(uncomp_qbs + recompute_qubits).intersection(instr.qubits):
                break
            previous_instructions.append(qc_new.data.pop(0))

        # Determine the the follow-up instructions.
        while len(qc_new.data):
            instr = qc_new.data[-1]
            if set(uncomp_qbs + recompute_qubits).intersection(instr.qubits):
                break
            follow_up_instructions.append(qc_new.data.pop(-1))

        follow_up_instructions = follow_up_instructions[::-1]

        # Build up the pdag
        pdag = PermeabilityGraph(qc_new)
        # Now come the steps of the Unqomp algorithm

        # We iterate over a reversed topological sort and successively
        # uncompute all relevant nodes.
        lin = list(nx.topological_sort(pdag))
        lin.reverse()

        for i in range(len(lin)):

            # Set an alias
            node = lin[i]

            # If the node is a terminator it needs no uncomputation
            if isinstance(node, TerminatorNode):
                continue

            # Get the target qubits
            # target qubits means the qubits the corresponding instruction is
            # operating either with X-permeability or neutral
            target_qubits = pdag.get_target_qubits(node)

            # If there is no intersection between the target qubits of the instruction
            # and the qubits to uncompute, this instruction needs no uncomputation
            if not set(target_qubits).intersection(uncomp_qbs):
                continue

            # This function checks if there is more than one allocation gate in
            # the upcoming chain of nodes. This prevents instructions which were
            # created due to previous recomputations to be included in further recomputations
            if detect_double_alloc(lin[i:], node.instr.qubits):
                continue

            # This case represents the neccesity to uncompute
            if set(target_qubits).issubset(uncomp_qbs) and node.instr and target_qubits:

                # The uncompute node function inserts an uncomputation node into the pdag
                # If a recomputation is required, no nodes are inserted and it returns True
                recompute = uncompute_node(pdag, node, uncomp_qbs, recompute_qubits)

                # If a recomputation is required, we call the algorithm with the approriate
                # recomputation qubits. Qubits that could require recomputation can only
                # be part of the controls
                if recompute:
                    return uncompute_qc(
                        qc,
                        uncomp_qbs
                        + list(
                            set(pdag.get_control_qubits(node)).intersection(
                                recompute_qubits
                            )
                        ),
                        recompute_qubits,
                    )
                continue

            # This case treats the possibility that the instruction is targeting qubits
            # which are not part of the uncomputation qubits.

            # This would is usually not uncomputable, however it could also imply
            # that these qubits have been uncomputed allready but gate_wrapped.
            # The gate_wrap decorator makes ancilla slots as non-permeable,
            # so these qubits appear as target qubits here.
            off_target_qubits = list(set(target_qubits) - set(uncomp_qbs))

            if off_target_qubits:

                non_uncomputable_qubits = []
                for qb in off_target_qubits:
                    if qb.allocated:
                        non_uncomputable_qubits.append(qb)

                if non_uncomputable_qubits:
                    raise Exception(
                        f'Uncomputation failed because gate "{node.instr.op.name}" needs to be uncomputed '
                        f"but is also targeting qubits {non_uncomputable_qubits} which are not up for uncomputation"
                    )
                else:
                    return uncompute_qc(
                        qc,
                        uncomp_qbs + off_target_qubits,
                        recompute_qubits=recompute_qubits,
                    )

        try:
            # Retrieve the uncomputed QuantumCircuit
            uncomputed_qc = pdag.to_qc()
        except nx.NetworkXUnfeasible:
            raise Exception("Cyclic dependency detected in DAG during uncomputation")

        # Insert the previous and follow_up instructions
        uncomputed_qc.data = (
            previous_instructions + uncomputed_qc.data + follow_up_instructions
        )

        # Return the uncomputed QuantumCircuit
        return uncomputed_qc


def uncompute_node(pdag, node, uncomp_qbs, recompute_qubits=[]):
    """
    Uncomputes a node in a given PermeabilityGraph (in-place) according to the
    Unqomp algorithm.

    Parameters
    ----------
    pdag : PermeabilityGraph
        The PermeabilityGraph to perform the uncomputation in.
    node : UnqompNode
        The node to uncompute.
    uncomp_qbs : list[Qubit]
        A list of Qubits that are supposed to be uncomputed.
    recompute_qubits : list[Qubit], optional
        A list of Qubit, where the algorithm is allowed to recompute them. The default is [].

    Raises
    ------
    Exception
        Tried to uncompute non-qfree instruction

    Returns
    -------
    bool
        A bool indicating whether a recomputation is required.

    """

    # Get the target qubits
    target_qubits = pdag.get_target_qubits(node)

    # In the Unqomp paper there is only one a* node.
    # However since we also allow multiple target operations, our case also
    # includes the possibility of having multiple a*

    # Retrieve the list of a* nodes
    a_star_n_list = [pdag.recent_node_dic[qb] for qb in set(target_qubits)]

    # Remove duplicates
    a_star_n_list = list(set(a_star_n_list))

    if not is_qfree(node.instr.op):
        raise Exception(f"Tried to uncompute non-qfree instruction {node.instr}")

    # Retrieve the list of nodes that are connected to node by a control edges.
    ctrls = pdag.get_control_nodes(node)

    # If the control node is already uncomputed, we use this node instead.
    # for i in range(len(ctrls)):
    #     if not ctrls[i].uncomputed_node is None:
    #         ctrls[i] = ctrls[i].uncomputed_node

    # Replace controlled operations by phase tolerant controlled operations
    if isinstance(node.instr.op, ControlledOperation):
        op = node.instr.op

        if op.method == "auto" and len(op.controls) < 5 or op.method == "gray":
            if len(op.controls) != 1:
                node.instr.op = PTControlledOperation(
                    op.base_operation,
                    num_ctrl_qubits=len(op.controls),
                    ctrl_state=op.ctrl_state,
                    method="gray_pt",
                )

    from qrisp.alg_primitives.logic_synthesis import LogicSynthGate

    # Replace results of logic synthesis by phase tolerant logic synthesis
    if isinstance(node.instr.op, LogicSynthGate):
        if node.instr.op.logic_synth_method == "gray":
            from qrisp import QuantumVariable

            tt = node.instr.op.tt

            input_qv = QuantumVariable(tt.bit_amount)
            output_qv = QuantumVariable(tt.shape[1], qs=input_qv.qs)

            tt.q_synth(input_qv, output_qv, method="gray_pt")

            temp = input_qv.qs.data[-1].op
            temp.name = node.instr.op.name

            node.instr.op = temp

    # Create the instruction for the new UnqompNode
    new_instr = node.instr.inverse()

    # Create the new UnqompNode
    reversed_node = InstructionNode(new_instr)

    # Link the uncomputed node to the original node
    node.uncomputed_node = reversed_node

    # Add the uncomputed node to the pdag
    pdag.add_node(reversed_node)

    reversed_node.value_layer = max([n.value_layer + 1 for n in a_star_n_list + ctrls])

    # Update the recent_node_dic with the reversed node
    for qb in target_qubits:
        pdag.recent_node_dic[qb] = reversed_node

    # We now connect the edges of the reversed node.

    # The first step is to connect the controls.
    # The idea here is to connect the reversed node to the
    # controls of the original node.

    # This dictionary will hold the control qubits of each control edge
    ctrl_qubit_dic = {}
    for c in ctrls:
        ctrl_qubit_dic[c] = pdag.get_edge_qubits(c, node)

    while len(ctrls):

        # Set alias
        c = ctrls.pop(0)

        # Get the qubits of the control edge
        control_edge_qubits = ctrl_qubit_dic[c]

        recomputation_required = set(control_edge_qubits).intersection(recompute_qubits)

        # This treats the case that the control edge is among the recomputable qubits
        # implying this node needs to be recomputed.
        if recomputation_required and pdag.has_edge(c, node):

            # If the required qubits for recomputation are not among the uncomputation
            # qubits, we cancel the function by returning True
            # The parent function will then start a new uncomputation attempt,
            # where the recompute qubits are among the uncomputation qubits.
            if not recomputation_required.issubset(uncomp_qbs):
                return True

            # Instead of connecting to the previos control node, we connect to the
            # "latest" control node that operated on that qubit.
            for qb in recomputation_required:
                ctrls.append(pdag.recent_node_dic[qb])
                ctrl_qubit_dic[pdag.recent_node_dic[qb]] = [qb]

            continue

        # We now add the anti_dependency edges starting at the reversed node

        # For this we iterate over the control qubits and check the streak
        # associated to that qubit
        for qb in control_edge_qubits:

            streak_children = pdag.get_streak_children(c, qb)

            # Iterate over the streak
            for streak_child in streak_children:

                # The streak is cancelled by the streak children of the streak_child
                for cancellation_node in pdag.get_streak_children(streak_child, qb):

                    pdag.add_edge(
                        reversed_node,
                        cancellation_node,
                        edge_type="anti_dependency",
                        qubits=[qb],
                    )

                    cancellation_node.value_layer = max(
                        cancellation_node.value_layer, reversed_node.value_layer + 1
                    )

        # Add the edge from the control node of the original node to the reversed node.
        pdag.add_edge(c, reversed_node, edge_type="Z", qubits=control_edge_qubits)

    # The next step is to connect the targets.
    # For that we use the a_star_n_list which represents all the nodes, that are
    # targetted by the operation
    for a_star_n in a_star_n_list:

        # We first add the edge to the qubits targetted by a_star_n

        # The qubits of this edge are the qubits which are:
        #   1. Part of the uncomputation qubits
        #   2. Targeting a_star_n
        #   3. Part of the instruction

        # Add the edge

        # We need to consider the possibility that we end a streak by adding this node
        # we therefore consider the nodes that use a_star_n as a control node

        # We now iterate over the targets of the node to connect the edges
        for qb in target_qubits:

            # These are the nodes which form a streak on that qubit
            streak_children = pdag.get_streak_children(a_star_n, qb)
            # Depending on how long the streak is, we have ton insert a TerminatorNode

            if len(streak_children) == 0:
                pdag.add_edge(a_star_n, reversed_node, edge_type="neutral", qubits=[qb])

            # If there is only one control, we can safely append a neutral edge to that control node
            elif len(streak_children) == 1:
                pdag.add_edge(
                    streak_children[0], reversed_node, edge_type="neutral", qubits=[qb]
                )

            # If there is a streak, we need to insert a terminator edge
            else:
                t_node = TerminatorNode(qb)
                t_node.value_layer = 0
                for v in streak_children:
                    pdag.add_edge(v, t_node, edge_type="anti_dependency", qubits=[qb])

                    if v.value_layer > t_node.value_layer:
                        t_node.value_layer = v.value_layer + 1

                # Add the edge
                pdag.add_edge(t_node, reversed_node, edge_type="neutral", qubits=[qb])

    return False


def detect_double_alloc(lin, qubits):
    qb_allocs = []
    for node in lin:
        if not node.instr:
            continue
        if node.instr.op.name == "qb_alloc" and node.instr.qubits[0] in qubits:
            if node.instr.qubits[0] in qb_allocs:
                return True
            else:
                qb_allocs.append(node.instr.qubits[0])

    return False



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from qrisp import QuantumCircuit
from qrisp.alg_primitives.arithmetic.adders.adder_tools import ammend_inpl_adder

# Implementation of the modular Cuccaro-Adder https://arxiv.org/pdf/quant-ph/0410184.pdf


def MAJ_gate():
    qc = QuantumCircuit(3)

    qc.cx(2, 1)
    qc.cx(2, 0)
    qc.mcx([0, 1], 2)

    result = qc.to_gate()

    result.name = "MAJ"

    return result


def UMA_gate(mode=2):
    qc = QuantumCircuit(3)
    if mode == 2:
        qc.mcx([0, 1], 2)
        qc.cx(2, 0)
        qc.cx(0, 1)

    if mode == 3:
        qc.x(1)
        qc.cx(0, 1)
        qc.mcx([0, 1], 2)
        qc.x(1)
        qc.cx(2, 0)
        qc.cx(2, 1)

    result = qc.to_gate()
    result.name = "UMA"
    return result


# Performs inplace addition on qv2, ie.
# qv1 += qv2


# TO-DO make this less circuity and more qrispy
def cuccaro_procedure(qs, qubit_list_1, qubit_list_2, output_qubit=None, carry_in=None):
    if len(qubit_list_1) != len(qubit_list_2):
        print(len(qubit_list_1))
        print(len(qubit_list_2))
        raise Exception(
            "Tried to call Cuccaro-procedure with qubit lists of unequal length"
        )

    from qrisp.core import QuantumVariable

    if carry_in is None:
        # Request ancilla Qubit
        ancilla = QuantumVariable(1)
    else:
        ancilla = [carry_in]

    # Prepare MAJ/UMA gate qubits
    slot_1_qbs = list(ancilla) + qubit_list_1[:-1]
    slot_2_qbs = qubit_list_2
    slot_3_qbs = qubit_list_1

    iterations = len(slot_1_qbs)

    # Perform 1st step of the modular addition section
    for i in range(iterations):
        qbits = [slot_1_qbs[i], slot_2_qbs[i], slot_3_qbs[i]]
        qs.append(MAJ_gate(), qbits)

    # Calculate output qbit
    if output_qubit:
        qs.cx(qbits[2], output_qubit)

    # Perform UMA iterations
    for i in range(iterations - 1, -1, -1):
        qbits = [slot_1_qbs[i], slot_2_qbs[i], slot_3_qbs[i]]

        qs.append(UMA_gate(), qbits)

    if carry_in is None:
        # Detete ancilla
        ancilla.delete()


def cuccaro_adder(a, b, c_in=None, c_out=None):
    """
    In-place adder function based on `this paper <https://arxiv.org/abs/quant-ph/0410184>`__
    Performs the addition

    ::

        b += a


    Parameters
    ----------
    a : int or QuantumVariable or list[Qubit]
        The value that should be added.
    b : QuantumVariable or list[Qubit]
        The value that should be modified in the in-place addition.
    c_in : Qubit, optional
        An optional carry in value. The default is None.
    c_out : Qubit, optional
        An optional carry out value. The default is None.

    Examples
    --------

    We add two integers:

    >>> from qrisp import QuantumFloat, cuccaro_adder
    >>> a = QuantumFloat(4)
    >>> b = QuantumFloat(4)
    >>> a[:] = 4
    >>> b[:] = 5
    >>> cuccaro_adder(a,b)
    >>> print(b)
    {9: 1.0}

    """

    cuccaro_procedure(b[0].qs(), a, b, carry_in=c_in, output_qubit=c_out)


temp = cuccaro_adder.__doc__
cuccaro_adder = ammend_inpl_adder(cuccaro_adder)
cuccaro_adder.__doc__ = temp



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import numpy as np
import threading

from qrisp.circuit.quantum_circuit import QuantumCircuit


class CompilationAccelerator:

    def __init__(self, xla_mode=2):
        self.xla_mode = xla_mode

    def __enter__(self):

        
        self.original_xla_mode = QuantumCircuit.xla_mode

        if threading.current_thread() is threading.main_thread():
            QuantumCircuit.xla_mode = self.xla_mode

    def __exit__(self, exception_type, exception_value, traceback):
        if threading.current_thread() is threading.main_thread():
            QuantumCircuit.xla_mode = self.original_xla_mode


fast_append = CompilationAccelerator



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""


from qrisp import h, QuantumFloat, multi_measurement, QuantumModulus, gphase
import numpy as np


def test_quantum_float_comparison():

    a = QuantumFloat(5, 2, signed=False)
    b = QuantumFloat(5, -1, signed=True)
    # Test all the operators
    for comp in ["eq", "neq", "lt", "gt", "leq", "geq"]:
        comparison_helper(a, b, comp)

    # Test specific constellations of QuantumFloat parameters
    a = QuantumFloat(3, -1, signed=True)
    b = QuantumFloat(5, 1, signed=True)
    comparison_helper(a, b, "lt")

    a = QuantumFloat(5, 2, signed=False)
    b = QuantumFloat(5, -1, signed=True)
    comparison_helper(a, b, "lt")

    a = QuantumFloat(4, -4, signed=False)
    b = QuantumFloat(4, 0, signed=True)
    comparison_helper(a, b, "lt")

    a = QuantumFloat(4, 1, signed=True)
    b = QuantumFloat(3, -1, signed=False)
    comparison_helper(a, b, "lt")

    # Test semi-classical comparisons
    a = QuantumFloat(4, signed=False)
    b = 4
    comparison_helper(a, b, "lt")
    comparison_helper(a, b, "gt")

    a = QuantumFloat(4, signed=True)
    b = 4
    comparison_helper(a, b, "lt")
    comparison_helper(a, b, "gt")

    a = QuantumFloat(4, -2, signed=False)
    b = 4
    comparison_helper(a, b, "lt")
    comparison_helper(a, b, "gt")

    a = QuantumFloat(8, 3, signed=False)
    b = 16
    comparison_helper(a, b, "lt")
    comparison_helper(a, b, "gt")


def test_quantum_modulus_comparison():
    
    a = QuantumModulus(5)
    b = QuantumModulus(5)
    
    for comp in ["eq", "neq", "lt", "gt", "leq", "geq"]:
        comparison_helper(a, b, comp)
        
    a = QuantumModulus(7)
    b = QuantumModulus(7)
    
    for comp in ["eq", "neq", "lt", "gt", "leq", "geq"]:
        comparison_helper(a, b, comp)
        
    a = QuantumModulus(13)
    b = QuantumModulus(13)
    
    for comp in ["eq", "neq", "lt", "gt", "leq", "geq"]:
        comparison_helper(a, b, comp)
        
    a = QuantumModulus(5)
    b = 3
    
    for comp in ["eq", "neq", "lt", "gt", "leq", "geq"]:
        comparison_helper(a, b, comp)
        
    a = QuantumModulus(7)
    b = 5
    
    for comp in ["eq", "neq", "lt", "gt", "leq", "geq"]:
        comparison_helper(a, b, comp)
        
    a = QuantumModulus(13)
    b = 7
    
    for comp in ["eq", "neq", "lt", "gt", "leq", "geq"]:
        comparison_helper(a, b, comp)
        
    
    # Test comparisons as ConditionalEnvironments
    
    qf1 = QuantumFloat(2)
    qf2 = QuantumFloat(2)

    h(qf1)
    h(qf2)

    with qf1==qf2:
        gphase(np.pi,qf1[0])

    qf1.qs.statevector()
    
    qf1 = QuantumFloat(2)
    qf2 = QuantumFloat(2)

    h(qf1)
    h(qf2)

    with qf1<qf2:
        gphase(np.pi,qf1[0])

    qf1.qs.statevector()

    

def comparison_helper(qf_0, qf_1, comp):
    qf_0 = qf_0.duplicate()
    h(qf_0)

    if isinstance(qf_1, (QuantumFloat, QuantumModulus)):
        qf_1 = qf_1.duplicate()
        h(qf_1)

    if comp == "eq":
        qf_res = qf_0 == qf_1
    elif comp == "neq":
        qf_res = qf_0 != qf_1
    elif comp == "leq":
        qf_res = qf_0 <= qf_1
    elif comp == "geq":
        qf_res = qf_0 >= qf_1
    elif comp == "lt":
        qf_res = qf_0 < qf_1
    elif comp == "gt":
        qf_res = qf_0 > qf_1

    if isinstance(qf_1, (QuantumFloat, QuantumModulus)):
        mes_res = multi_measurement([qf_0, qf_1, qf_res])
        for a, b, c in mes_res.keys():
            
            if a is np.nan or b is np.nan:
                continue
            
            print(a, b, c)
            
            
            if comp == "eq":
                assert (a == b) == c
            elif comp == "neq":
                assert (a != b) == c
            elif comp == "leq":
                assert (a <= b) == c
            elif comp == "geq":
                assert (a >= b) == c
            elif comp == "lt":
                assert (a < b) == c
            elif comp == "gt":
                assert (a > b) == c

        # Test correct phase behavior7
        statevector = qf_res.qs.statevector("array")
        angles = np.angle(
            statevector[
                np.abs(statevector) > 1 / 2 ** ((qf_0.size + qf_1.size) / 2)
            ]
        )
        assert np.sum(np.abs(angles)) < 0.1
    else:
        mes_res = multi_measurement([qf_0, qf_res])
        for a, c in mes_res.keys():
            
            if a is np.nan:
                continue
            
            b = qf_1
            print(a, b, c)
            if comp == "eq":
                assert (a == b) == c
            elif comp == "neq":
                assert (a != b) == c
            elif comp == "leq":
                assert (a <= b) == c
            elif comp == "geq":
                assert (a >= b) == c
            elif comp == "lt":
                assert (a < b) == c
            elif comp == "gt":
                assert (a > b) == c


"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from functools import lru_cache

from jax import make_jaxpr
from jax.core import JaxprEqn, ClosedJaxpr
from jax.lax import add_p, sub_p, while_loop

from qrisp.jasp.primitives import AbstractQuantumCircuit, OperationPrimitive


def copy_jaxpr_eqn(eqn):
    return JaxprEqn(
        primitive=eqn.primitive,
        invars=list(eqn.invars),
        outvars=list(eqn.outvars),
        params=dict(eqn.params),
        source_info=eqn.source_info,
        effects=eqn.effects,
    )


def invert_eqn(eqn):
    """
    Receives and equation that describes either an operation or a pjit primitive
    and returns an equation that describes the inverse.

    Parameters
    ----------
    eqn : jax.core.JaxprEqn
        The equation to be inverted.

    Returns
    -------
    inverted_eqn
        The equation with inverted operation.

    """

    if eqn.primitive.name == "pjit":
        params = dict(eqn.params)
        params["jaxpr"] = ClosedJaxpr(
            (eqn.params["jaxpr"].jaxpr).inverse(), eqn.params["jaxpr"].consts
        )

        name = params["name"]
        if name[-3:] == "_dg":
            params["name"] = params["name"][:-3]
        elif name not in ["gidney_mcx", "gidney_mcx_inv"]:
            params["name"] += "_dg"
        elif name == "gidney_mcx":
            params["name"] = "gidney_mcx_inv"
        elif name == "gidney_mcx_inv":
            params["name"] = "gidney_mcx"

        primitive = eqn.primitive
    elif eqn.primitive.name == "while":
        return invert_loop_eqn(eqn)
    elif eqn.primitive.name == "cond":
        params = dict(eqn.params)

        branches = params["branches"]

        new_branch_list = []

        for i in range(len(branches)):
            new_branch_list.append(
                ClosedJaxpr((branches[i].jaxpr).inverse(), branches[i].consts)
            )

        params["branches"] = tuple(new_branch_list)

        primitive = eqn.primitive

    else:
        params = dict(eqn.params)
        primitive = eqn.primitive.inverse()

    return JaxprEqn(
        primitive=primitive,
        invars=list(eqn.invars),
        outvars=list(eqn.outvars),
        params=params,
        source_info=eqn.source_info,
        effects=eqn.effects,
    )


@lru_cache(int(1e5))
def invert_jaspr(jaspr):
    """
    Takes a Jaspr and returns a Jaspr, which performs the inverted quantum operation

    Parameters
    ----------
    jaspr : qrisp.jasp.Jaspr
        The Jaspr to be inverted.

    Returns
    -------
    inverted_jaspr : qrisp.jasp.Jaspr
        The inverted/daggered Jaspr.

    """

    # Flatten all environments in the jaspr
    jaspr = jaspr.flatten_environments()
    # We separate the equations into classes where one executes Operations and
    # the one that doesn't execute Operations
    if jaspr.inv_jaspr is not None:
        return jaspr.inv_jaspr

    op_eqs = []
    non_op_eqs = []
    deletions = []

    # Since the Operation equations require as inputs only qubit object and a QuantumCircuit
    # we achieve our goal by pulling all the non-Operation equations to the front
    # and the Operation equations to the back.

    for eqn in jaspr.eqns:
        if isinstance(eqn.primitive, OperationPrimitive) or (
            (eqn.primitive.name in ["pjit", "while", "cond"])
            and len(eqn.invars)
            and isinstance(eqn.invars[-1].aval, AbstractQuantumCircuit)
        ):
            # Insert the inverted equation at the front
            op_eqs.insert(0, invert_eqn(eqn))
        elif eqn.primitive.name == "jasp.measure":
            raise Exception("Tried to invert a jaspr containing a measurement")
        elif eqn.primitive.name == "jasp.delete_qubits":
            deletions.append(eqn)
        else:
            non_op_eqs.append(eqn)

    # Finally, we need to make sure the Order of QuantumCircuit I/O is also reversed.
    n = len(op_eqs)
    if n == 0:
        return jaspr

    for i in range(n // 2):

        for j in range(len(op_eqs[i].invars)):
            if isinstance(op_eqs[i].invars[j].aval, AbstractQuantumCircuit):
                break
        else:
            raise

        for k in range(len(op_eqs[n - i - 1].invars)):
            if isinstance(op_eqs[n - i - 1].invars[k].aval, AbstractQuantumCircuit):
                break
        else:
            raise

        op_eqs[i].invars[j], op_eqs[n - i - 1].invars[k] = (
            op_eqs[n - i - 1].invars[k],
            op_eqs[i].invars[j],
        )

        for j in range(len(op_eqs[i].outvars)):
            if isinstance(op_eqs[i].outvars[j].aval, AbstractQuantumCircuit):
                break
        else:
            raise

        for k in range(len(op_eqs[n - i - 1].outvars)):
            if isinstance(op_eqs[n - i - 1].outvars[k].aval, AbstractQuantumCircuit):
                break
        else:
            raise

        op_eqs[i].outvars[j], op_eqs[n - i - 1].outvars[k] = (
            op_eqs[n - i - 1].outvars[k],
            op_eqs[i].outvars[j],
        )

    from qrisp.jasp import Jaspr

    # Find the last AbstractQuantumCircuit variable

    for j in range(len(op_eqs[-1].outvars)):
        if isinstance(op_eqs[-1].outvars[j].aval, AbstractQuantumCircuit):
            last_abs_qc = op_eqs[-1].outvars[j]
            break

    if len(deletions):
        first_deletion = copy_jaxpr_eqn(deletions[0])
        first_deletion.invars[-1] = last_abs_qc
        last_abs_qc = deletions[-1].outvars[-1]

    res = Jaspr(
        constvars=jaspr.constvars,
        invars=list(jaspr.invars),
        outvars=jaspr.outvars[:-1] + [last_abs_qc],
        eqns=non_op_eqs + op_eqs + deletions,
    )

    if jaspr.ctrl_jaspr:
        res.ctrl_jaspr = jaspr.ctrl_jaspr.inverse()

    return res


# The following two functions manage the logic behind inverting a loop.
# For this we have to consider two aspects.

# 1. The loop body contains the loop index incrementation. This incrementation
#    needs to be reversed to a decrementation. Furthermore the quantum part of
#    the loop body needs to be inverted.

# 2. The loop index initialization and the threshold need to be switched and
#    the comparison to determine the break condition needs to be adjusted.


def invert_loop_body(jaxpr):
    # This function treats the loop body

    # This list will contain the equations with the loop index decrementation
    new_eqn_list = list(jaxpr.eqns)

    # Find the incrementation equation. For this we identify the equation,
    # which updates the loop index

    loop_index = jaxpr.jaxpr.outvars[1]
    for i in range(len(new_eqn_list))[::-1]:
        if loop_index == new_eqn_list[i].outvars[0]:
            break
    increment_eqn_index = int(i)
    increment_eqn = new_eqn_list[increment_eqn_index]

    # Change the primitive
    if increment_eqn.primitive is add_p:
        new_primitive = sub_p
    else:
        new_primitive = add_p

    try:
        if increment_eqn.invars[1].val != 1:
            raise
    except:
        raise Exception(
            "Dynamic loop inversion is only supported for loops the have step size 1."
        )

    # Create the decrement equation
    decrement_eqn = JaxprEqn(
        primitive=new_primitive,
        invars=list(increment_eqn.invars),
        outvars=list(increment_eqn.outvars),
        params=increment_eqn.params,
        source_info=increment_eqn.source_info,
        effects=increment_eqn.effects,
    )
    new_eqn_list[increment_eqn_index] = decrement_eqn

    # Create the new Jaspr
    from qrisp.jasp import Jaspr

    new_jaspr = Jaspr(jaxpr.jaxpr).update_eqns(new_eqn_list)

    # Quantum inversion
    res_jaspr = new_jaspr.inverse()

    return res_jaspr


# This function performs the above mentioned step 2 to treat the loop primitive
def invert_loop_eqn(eqn):

    # Process the loop body
    body_jaxpr = eqn.params["body_jaxpr"]
    inv_loop_body = invert_loop_body(body_jaxpr)

    def body_fun(val):
        return inv_loop_body.eval(*val)

    # Process the loop cancelation
    cond_jaxpr = eqn.params["cond_jaxpr"]

    def cond_fun(val):
        if cond_jaxpr.eqns[0].primitive.name == "ge":
            return val[1] <= val[0]
        else:
            return val[1] >= val[0]

    # Create the new equation by tracing the while loop
    def tracing_function(*args):
        return while_loop(cond_fun, body_fun, tuple(args))

    jaxpr = make_jaxpr(tracing_function)(*[var.aval for var in eqn.invars])
    new_eqn = jaxpr.eqns[0]

    # The new invars should have initial loop index at loop threshold switched.
    # The loop initialization is located at invars[1] and the threshold at invars[0]
    invars = eqn.invars
    new_invars = list(invars)
    new_invars[1], new_invars[0] = new_invars[0], new_invars[1]

    # Create the Equation
    res = JaxprEqn(
        primitive=new_eqn.primitive,
        invars=new_invars,
        outvars=list(eqn.outvars),
        params=new_eqn.params,
        source_info=eqn.source_info,
        effects=eqn.effects,
    )

    return res



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import jax
import jax.numpy as jnp

from qrisp.core import recursive_qv_search, recursive_qa_search

from qrisp.jasp.primitives import AbstractQuantumCircuit
from qrisp.jasp.tracing_logic import TracingQuantumSession, check_for_tracing_mode


def qache(*func, **kwargs):
    """
    This decorator allows you to mark a function as "reusable". Reusable here means
    that the jasp expression of this function will be cached and reused in the next
    calls (if the function is called with the same signature).

    A qached function therefore has to be traced by the Python interpreter only once
    and after that the function can be called without any Python-interpreter induced
    delay. This can significantly speed up the compilation process.

    Using the ``qache`` decorator not only improves the compilation speed but also
    enables the compiler to speed up transformation processes.

    .. warning::

        Two important rules apply to the ``qache`` decorator to adhere to the
        functional programming paradigm.

        * It is illegal to have a qached function return a QuantumVariable that has been passed as an argument to the function.
        * It is illegal to modify traced attributes of QuantumVariables that have been passed as an argument to the function.

        See the examples section for representatives of these cases.

    Parameters
    ----------
    func : callable
        The function to be qached.

    Returns
    -------
    qached_function : callable
        A function that will be traced on it's first execution and retrieved from
        the cache in any other call.

    Examples
    --------

    We create a simple function that is qached. To simulate an expensive compilation
    task we insert a ``time.sleep`` command.

    ::

        import time
        from qrisp import *
        from qrisp.jasp import qache

        @qache
        def inner_function(qv):
            h(qv[0])
            cx(qv[0], qv[1])
            res_bl = measure(qv[0])

            # Simulate demanding compilation procedure by calling
            time.sleep(1)

            return res_bl

        def main():
            a = QuantumVariable(2)
            b = QuantumFloat(2)

            bl_0 = inner_function(a)
            bl_1 = inner_function(b)
            bl_2 = inner_function(a)
            bl_3 = inner_function(b)

            return bl_0 & bl_1 & bl_2 & bl_3

        # Measure the time required for tracing
        t0 = time.time()
        jaspr = make_jaspr(main)()
        print(time.time() - t0) # 2.0225703716278076

    Even though ``inner_function`` has been called 4 times, we only see a delay of 2 seconds.
    This is because the function has been called with two different quantum types, implying it
    has been traced twice and recalled from the cache twice. We take a look at the :ref:`jaspr`.

    >>> print(jaspr)
    let inner_function = { lambda ; a:QuantumCircuit b:QubitArray. let
        c:Qubit = jasp.get_qubit b 0
        d:QuantumCircuit = jasp.h a c
        e:Qubit = jasp.get_qubit b 1
        f:QuantumCircuit = jasp.cx d c e
        g:QuantumCircuit h:bool[] = jasp.measure f c
      in (g, h) } in
    let inner_function1 = { lambda ; i:QuantumCircuit j:QubitArray k:i32[] l:bool[]. let
        m:Qubit = jasp.get_qubit j 0
        n:QuantumCircuit = jasp.h i m
        o:Qubit = jasp.get_qubit j 1
        p:QuantumCircuit = jasp.cx n m o
        q:QuantumCircuit r:bool[] = jasp.measure p m
      in (q, r) } in
    { lambda ; s:QuantumCircuit. let
        t:QuantumCircuit u:QubitArray = jasp.create_qubits s 2
        v:QuantumCircuit w:QubitArray = jasp.create_qubits t 2
        x:QuantumCircuit y:bool[] = pjit[name=inner_function jaxpr=inner_function] v
          u
        z:QuantumCircuit ba:bool[] = pjit[name=inner_function jaxpr=inner_function1] x
          w 0 False
        bb:QuantumCircuit bc:bool[] = pjit[name=inner_function jaxpr=inner_function] z
          u
        bd:QuantumCircuit be:bool[] = pjit[
          name=inner_function
          jaxpr=inner_function1
        ] bb w 0 False
        bf:bool[] = and y ba
        bg:bool[] = and bf bc
        bh:bool[] = and bg be
        bi:QuantumCircuit = jasp.reset bd u
        bj:QuantumCircuit = jasp.delete_qubits bi u
      in (bj, bh) }

    As expected, we see three different function definitions:

    * The first one describes ``inner_function`` called with a :ref:`QuantumVariable`. For this kind of signature only the ``QubitArray`` is required.
    * The second one describes ``inner_function`` called with :ref:`QuantumFloat`. Additionally to the ``QubitArray``, the ``.exponent`` and ``.signed`` attribute are also passed to the function.
    * The third function definition is ``outer_function``, which calls the previously defined functions.

    **Illegal functions**

    We will now demonstrate what type of functions can not be qached.

    ::

        @qache
        def inner_function(qv):
            h(qv[0])
            return qv

        @jaspify
        def main():
            qf_0 = QuantumFloat(2)
            qf_1 = inner_function(qf_0)
            return measure(qf_1)

        main()
        # Yields: Exception: Found parameter QuantumVariable within returned results

    ``inner_function`` returns a :ref:`QuantumVariable` that has been passed as an
    argument and can therefore not be qached.

    The second case of an illegal functions is a function that tries to modify
    a traced attribute of a ``QuantumVariable`` that has been passed as an argument.
    A traced attribute is for instance the ``exponent`` attribute of :ref:`QuantumFloat`.

    ::

        @qache
        def inner_function(qf):
            qf.exponent += 1

        @jaspify
        def main():
            qf = QuantumFloat(2)
            inner_function(qf)

        main()
        # Yields: Exception: Found in-place parameter modification of QuantumVariable qf

    """

    if len(kwargs) and len(func) == 0:
        return lambda x: qache_helper(x, kwargs)
    elif len(kwargs) and len(func):
        return qache_helper(func[0], kwargs)
    else:
        return qache_helper(func[0], {})


# temp_list = [False]
def qache_helper(func, jax_kwargs):

    # To achieve the desired behavior we leverage the Jax inbuild caching mechanism.
    # This feature can be used by calling a jitted function in a tracing context.
    # To cache the function we therefore simply need to wrap it with jit and
    # it will be properly cached.

    # if func.__name__ == "jasp_qq_gidney_adder":
    # if temp_list[0]:
    # raise
    # temp_list[0] = True

    # There are however some more things to consider.

    # The Qrisp function doesn't have the AbstractQuantumCircuit object (which is carried by
    # the tracing QuantumSession) in the signature.

    # To make jax properly treat this, we modify the function signature

    # This function performs the input function but also has the AbstractQuantumCircuit
    # in the signature.
    def ammended_function(*args, **kwargs):

        abs_qc = kwargs[10*"~"]
        del kwargs[10*"~"]

        # Set the given AbstractQuantumCircuit as the
        # one carried by the tracing QuantumSession
        abs_qs = TracingQuantumSession.get_instance()
        abs_qs.abs_qc = abs_qc

        # We now iterate through the QuantumVariables of the signature to perform two steps:
        # 1. The QuantumVariables from the signature went through a flatten/unflattening process.
        # The unflattening creates a copy of the QuantumVariable object, which is however not
        # registered in any QuantumSession. We therefore need to register them.
        # 2. To prevent the user from performing any in-place modifications of traced QuantumVariable
        # attributes, we collect the tracers to compare them after the function has concluded.
        arg_qvs = recursive_qv_search(args)
        arg_qvs += [qa.qtype for qa in recursive_qa_search(args)]

        flattened_qvs = []
        for qv in arg_qvs:
            abs_qs.register_qv(qv, None)
            flattened_qvs.extend(list(flatten_qv(qv)[0]))

        # Execute the function
        res = func(*args, **kwargs)

        res_qvs = recursive_qv_search(res)

        # It is not legal to return a QuantumVariable that was already given in the parameters.
        if set([hash(qv) for qv in res_qvs]).intersection([hash(qv) for qv in arg_qvs]):
            raise Exception("Found parameter QuantumVariable within returned results")

        res_qvs += [qa.qtype for qa in recursive_qa_search(res)]

        # Check whether there have been in-place modifications of traced attributes of QuantumVariables.
        for qv in arg_qvs:
            flat_qv = list(flatten_qv(qv)[0])
            for i in range(len(flat_qv)):
                if not flat_qv[i] is flattened_qvs.pop(0):
                    raise Exception(
                        f"Found in-place parameter modification of QuantumVariable {qv.name}"
                    )

        abs_qs.garbage_collection(arg_qvs + res_qvs)

        new_abs_qc = abs_qs.abs_qc
        # Return the result and the result AbstractQuantumCircuit.
        return res, new_abs_qc

    # Modify the name of the ammended function to reflect the input
    ammended_function.__name__ = func.__name__
    # Wrap in jax.jit
    ammended_function = jax.jit(ammended_function, **jax_kwargs)

    from qrisp.jasp.tracing_logic import flatten_qv

    # We now prepare the return function
    def return_function(*args, **kwargs):

        # If we are not in tracing mode, simply execute the function
        if not check_for_tracing_mode():
            return func(*args, **kwargs)

        # Get the AbstractQuantumCircuit for tracing
        abs_qs = TracingQuantumSession.get_instance()
        abs_qs.start_tracing(abs_qs.abs_qc)

        # Make sure literals are 32 bit
        args = list(args)
        # for i in range(len(args)):
        #     if isinstance(args[i], bool):
        #         args[i] = jnp.array(args[i], dtype = jnp.bool)
        #     elif isinstance(args[i], int):
        #         args[i] = jnp.array(args[i], dtype = jnp.int64)
        #     elif isinstance(args[i], float):
        #         args[i] = jnp.array(args[i], dtype = jnp.float64)
        #     elif isinstance(args[i], complex):
        #         args[i] = jnp.array(args[i], dtype = jnp.complex)

        # Excecute the function
        ammended_kwargs = dict(kwargs)
        ammended_kwargs[10*"~"] = abs_qs.abs_qc
        try:
            res, abs_qc_new = ammended_function(*args, **ammended_kwargs)
        except Exception as e:
            abs_qs.conclude_tracing()
            raise e

        abs_qs.conclude_tracing()

        # Convert the jaxpr from the traced equation in to a Jaspr
        from qrisp.jasp import Jaspr

        eqn = jax._src.core.thread_local_state.trace_state.trace_stack.dynamic.jaxpr_stack[
            0
        ].eqns[
            -1
        ]
        jaxpr = eqn.params["jaxpr"].jaxpr

        if not isinstance(eqn.invars[-1].aval, AbstractQuantumCircuit):
            for i in range(len(eqn.invars)):
                if isinstance(eqn.invars[i].aval, AbstractQuantumCircuit):
                    eqn.invars[-1], eqn.invars[i] = eqn.invars[i], eqn.invars[-1]
                    break
        if not isinstance(jaxpr.invars[-1].aval, AbstractQuantumCircuit):
            for i in range(len(jaxpr.invars)):
                if isinstance(jaxpr.invars[i].aval, AbstractQuantumCircuit):
                    jaxpr.invars[-1], jaxpr.invars[i] = (
                        jaxpr.invars[i],
                        jaxpr.invars[-1],
                    )
                    break

        eqn.params["jaxpr"] = jax.core.ClosedJaxpr(
            Jaspr.from_cache(jaxpr), eqn.params["jaxpr"].consts
        )

        # Update the AbstractQuantumCircuit of the TracingQuantumSession
        abs_qs.abs_qc = abs_qc_new

        # The QuantumVariables from the result went through a flatten/unflattening cycly.
        # The unflattening creates a new QuantumVariable object, that is however not yet
        # registered in any QuantumSession. We register these in the current QuantumSession.
        for qv in recursive_qv_search(res):
            abs_qs.register_qv(qv, None)

        # Return the result.
        return res

    return return_function



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import numpy as xp
import os

try:
    float_thresh = os.environ["QRISP_SIMULATOR_FLOAT_THRESH"]
except KeyError:
    float_thresh = 1e-5
float_thresh = xp.float32(float_thresh)

try:
    cutoff_ratio = os.environ["QRISP_SIMULATOR_CUTOFF_RATIO"]
except KeyError:
    cutoff_ratio = 2e-4
cutoff_ratio = xp.float32(cutoff_ratio)

try:
    sparsification_rate = os.environ["QRISP_SIMULATOR_SPARSIFICATION_RATE"]
except KeyError:
    sparsification_rate = 0.4
sparsification_rate = xp.float32(sparsification_rate)


"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from jax.core import AbstractValue, Primitive, raise_to_shaped_mappings


class AbstractQubit(AbstractValue):

    def __repr__(self):
        return "Qubit"

    def __hash__(self):
        return hash(type(self))

    def __eq__(self, other):
        if not isinstance(other, AbstractQubit):
            return False
        return isinstance(other, AbstractQubit)

    def _add(self, a, b):
        from qrisp.jasp import fuse_qb_array, DynamicQubitArray

        if isinstance(b, DynamicQubitArray):
            b = b.tracer
        return DynamicQubitArray(fuse_qb_array(a, b))


raise_to_shaped_mappings[AbstractQubit] = lambda aval, _: aval
